from unittest import mock

import pytest
from typeguard import TypeCheckError

from bec_lib import messages
from bec_lib.device import (
    AdjustableMixin,
    ComputedSignal,
    Device,
    DeviceBase,
    Positioner,
    RPCError,
    Signal,
    Status,
)
from bec_lib.devicemanager import DeviceContainer, DeviceManagerBase
from bec_lib.endpoints import MessageEndpoints
from bec_lib.tests.utils import ConnectorMock, get_device_info_mock

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access


@pytest.fixture(name="dev")
def fixture_dev(bec_client_mock):
    yield bec_client_mock.device_manager.devices


def test_nested_device_root(dev):
    assert dev.dyn_signals.name == "dyn_signals"
    assert dev.dyn_signals.messages.name == "messages"
    assert dev.dyn_signals.root == dev.dyn_signals
    assert dev.dyn_signals.messages.root == dev.dyn_signals


def test_read(dev):
    with mock.patch.object(dev.samx.root.parent.connector, "get") as mock_get:
        mock_get.return_value = messages.DeviceMessage(
            signals={
                "samx": {"value": 0, "timestamp": 1701105880.1711318},
                "samx_setpoint": {"value": 0, "timestamp": 1701105880.1693492},
                "samx_motor_is_moving": {"value": 0, "timestamp": 1701105880.16935},
            },
            metadata={"scan_id": "scan_id", "scan_type": "scan_type"},
        )
        res = dev.samx.read()
        mock_get.assert_called_once_with(MessageEndpoints.device_readback("samx"))
        assert res == {
            "samx": {"value": 0, "timestamp": 1701105880.1711318},
            "samx_setpoint": {"value": 0, "timestamp": 1701105880.1693492},
            "samx_motor_is_moving": {"value": 0, "timestamp": 1701105880.16935},
        }


def test_read_filtered_hints(dev):
    with mock.patch.object(dev.samx.root.parent.connector, "get") as mock_get:
        mock_get.return_value = messages.DeviceMessage(
            signals={
                "samx": {"value": 0, "timestamp": 1701105880.1711318},
                "samx_setpoint": {"value": 0, "timestamp": 1701105880.1693492},
                "samx_motor_is_moving": {"value": 0, "timestamp": 1701105880.16935},
            },
            metadata={"scan_id": "scan_id", "scan_type": "scan_type"},
        )
        res = dev.samx.read(filter_to_hints=True)
        mock_get.assert_called_once_with(MessageEndpoints.device_readback("samx"))
        assert res == {"samx": {"value": 0, "timestamp": 1701105880.1711318}}


def test_read_use_read(dev):
    with mock.patch.object(dev.samx.root.parent.connector, "get") as mock_get:
        data = {
            "samx": {"value": 0, "timestamp": 1701105880.1711318},
            "samx_setpoint": {"value": 0, "timestamp": 1701105880.1693492},
            "samx_motor_is_moving": {"value": 0, "timestamp": 1701105880.16935},
        }
        mock_get.return_value = messages.DeviceMessage(
            signals=data, metadata={"scan_id": "scan_id", "scan_type": "scan_type"}
        )
        res = dev.samx.read(use_readback=False)
        mock_get.assert_called_once_with(MessageEndpoints.device_read("samx"))
        assert res == data


def test_read_nested_device(dev):
    with mock.patch.object(dev.dyn_signals.root.parent.connector, "get") as mock_get:
        data = {
            "dyn_signals_messages_message1": {"value": 0, "timestamp": 1701105880.0716832},
            "dyn_signals_messages_message2": {"value": 0, "timestamp": 1701105880.071722},
            "dyn_signals_messages_message3": {"value": 0, "timestamp": 1701105880.071739},
            "dyn_signals_messages_message4": {"value": 0, "timestamp": 1701105880.071753},
            "dyn_signals_messages_message5": {"value": 0, "timestamp": 1701105880.071766},
        }
        mock_get.return_value = messages.DeviceMessage(
            signals=data, metadata={"scan_id": "scan_id", "scan_type": "scan_type"}
        )
        res = dev.dyn_signals.messages.read()
        mock_get.assert_called_once_with(MessageEndpoints.device_readback("dyn_signals"))
        assert res == data


@pytest.mark.parametrize(
    "kind,cached", [("normal", True), ("hinted", True), ("config", False), ("omitted", False)]
)
def test_read_kind_hinted(dev, kind, cached):
    with mock.patch.object(dev.samx.readback, "_run") as mock_run:
        with mock.patch.object(dev.samx.root.parent.connector, "get") as mock_get:
            data = {
                "samx": {"value": 0, "timestamp": 1701105880.1711318},
                "samx_setpoint": {"value": 0, "timestamp": 1701105880.1693492},
                "samx_motor_is_moving": {"value": 0, "timestamp": 1701105880.16935},
            }
            mock_get.return_value = messages.DeviceMessage(
                signals=data, metadata={"scan_id": "scan_id", "scan_type": "scan_type"}
            )
            dev.samx.readback._signal_info["kind_str"] = f"Kind.{kind}"
            res = dev.samx.readback.read(cached=cached)
            if cached:
                mock_get.assert_called_once_with(MessageEndpoints.device_readback("samx"))
                mock_run.assert_not_called()
                assert res == {"samx": {"value": 0, "timestamp": 1701105880.1711318}}
            else:
                mock_run.assert_called_once_with(cached=False, fcn=dev.samx.readback.read)
                mock_get.assert_not_called()


@pytest.mark.parametrize(
    "is_signal,is_config_signal,method",
    [
        (True, False, "read"),
        (False, True, "read_configuration"),
        (False, False, "read_configuration"),
    ],
)
def test_read_configuration_not_cached(dev, is_signal, is_config_signal, method):
    with mock.patch.object(
        dev.samx.readback, "_get_rpc_signal_info", return_value=(is_signal, is_config_signal, False)
    ):
        with mock.patch.object(dev.samx.readback, "_run") as mock_run:
            dev.samx.readback.read_configuration(cached=False)
            mock_run.assert_called_once_with(cached=False, fcn=getattr(dev.samx.readback, method))


@pytest.mark.parametrize(
    "is_signal,is_config_signal,method",
    [(True, False, "read"), (False, True, "redis"), (False, False, "redis")],
)
def test_read_configuration_cached(dev, is_signal, is_config_signal, method):
    with mock.patch.object(
        dev.samx.readback, "_get_rpc_signal_info", return_value=(is_signal, is_config_signal, True)
    ):
        with mock.patch.object(dev.samx.root.parent.connector, "get") as mock_get:
            mock_get.return_value = messages.DeviceMessage(
                signals={
                    "samx": {"value": 0, "timestamp": 1701105880.1711318},
                    "samx_setpoint": {"value": 0, "timestamp": 1701105880.1693492},
                    "samx_motor_is_moving": {"value": 0, "timestamp": 1701105880.16935},
                },
                metadata={"scan_id": "scan_id", "scan_type": "scan_type"},
            )
            with mock.patch.object(dev.samx.readback, "read") as mock_read:
                dev.samx.readback.read_configuration(cached=True)
                if method == "redis":
                    mock_get.assert_called_once_with(
                        MessageEndpoints.device_read_configuration("samx")
                    )
                    mock_read.assert_not_called()
                else:
                    mock_read.assert_called_once_with(cached=True)
                    mock_get.assert_not_called()


def test_run_rpc_call(dev):
    with mock.patch.object(dev.samx.setpoint, "_get_rpc_response") as mock_rpc:
        dev.samx.setpoint.set(1)
        mock_rpc.assert_called_once()


def test_set_calls_rpc(dev):
    with mock.patch.object(dev.samx.setpoint, "_run_rpc_call") as mock_rpc:
        dev.samx.setpoint.set(1)
        mock_rpc.assert_called_once_with("samx", "setpoint.set", 1)


def test_put_calls_set_if_wait(dev):
    with mock.patch.object(dev.samx.setpoint, "_run_rpc_call") as mock_rpc:
        dev.samx.setpoint.put(1, wait=True)
        mock_rpc.assert_called_once_with("samx", "setpoint.set", 1)


def test_put_calls_rpc(dev):
    with mock.patch.object(dev.samx.setpoint, "_run_rpc_call") as mock_rpc:
        dev.samx.setpoint.put(1)
        mock_rpc.assert_called_once_with("samx", "setpoint.put", 1)


def test_get_rpc_func_name_decorator(dev):
    with mock.patch.object(dev.samx.setpoint, "_run_rpc_call") as mock_rpc:
        dev.samx.setpoint.set(1)
        mock_rpc.assert_called_once_with("samx", "setpoint.set", 1)


def test_get_rpc_func_name_read(dev):
    with mock.patch.object(dev.samx, "_run_rpc_call") as mock_rpc:
        dev.samx.read(cached=False)
        mock_rpc.assert_called_once_with("samx", "read")


@pytest.mark.parametrize(
    "kind,cached", [("normal", True), ("hinted", True), ("config", False), ("omitted", False)]
)
def test_get_rpc_func_name_readback_get(dev, kind, cached):
    with mock.patch.object(dev.samx.readback, "_run") as mock_rpc:
        with mock.patch.object(dev.samx.root.parent.connector, "get") as mock_get:
            mock_get.return_value = messages.DeviceMessage(
                signals={
                    "samx": {"value": 0, "timestamp": 1701105880.1711318},
                    "samx_setpoint": {"value": 0, "timestamp": 1701105880.1693492},
                    "samx_motor_is_moving": {"value": 0, "timestamp": 1701105880.16935},
                },
                metadata={"scan_id": "scan_id", "scan_type": "scan_type"},
            )
            dev.samx.readback._signal_info["kind_str"] = f"Kind.{kind}"
            dev.samx.readback.get(cached=cached)
            if cached:
                mock_get.assert_called_once_with(MessageEndpoints.device_readback("samx"))
                mock_rpc.assert_not_called()
            else:
                mock_rpc.assert_called_once_with(cached=False, fcn=dev.samx.readback.get)
                mock_get.assert_not_called()


def test_get_rpc_func_name_nested(dev):
    with mock.patch.object(
        dev.samx._custom_rpc_methods["dummy_controller"]._custom_rpc_methods["_func_with_args"],
        "_run_rpc_call",
    ) as mock_rpc:
        dev.samx.dummy_controller._func_with_args(1, 2)
        mock_rpc.assert_called_once_with("samx", "dummy_controller._func_with_args", 1, 2)


def test_handle_rpc_response(dev):
    msg = messages.DeviceRPCMessage(device="samx", return_val=1, out="done", success=True)
    assert dev.samx._handle_rpc_response(msg) == 1


def test_handle_rpc_response_returns_status(dev, bec_client_mock):
    msg = messages.DeviceRPCMessage(
        device="samx", return_val={"type": "status", "RID": "request_id"}, out="done", success=True
    )
    assert dev.samx._handle_rpc_response(msg) == Status(
        bec_client_mock.device_manager, "request_id"
    )


def test_handle_rpc_response_raises(dev):
    msg = messages.DeviceRPCMessage(
        device="samx",
        return_val={"type": "status", "RID": "request_id"},
        out={
            "msg": "Didn't work...",
            "traceback": "Traceback (most recent call last):",
            "error": "error",
        },
        success=False,
    )
    with pytest.raises(RPCError):
        dev.samx._handle_rpc_response(msg)


def test_handle_rpc_response_returns_dict(dev):
    msg = messages.DeviceRPCMessage(device="samx", return_val={"a": "b"}, out="done", success=True)
    assert dev.samx._handle_rpc_response(msg) == {"a": "b"}


def test_run_rpc_call_calls_stop_on_keyboardinterrupt(dev):
    with mock.patch.object(dev.samx.setpoint, "_prepare_rpc_msg") as mock_rpc:
        mock_rpc.side_effect = [KeyboardInterrupt]
        with pytest.raises(RPCError):
            with mock.patch.object(dev.samx, "stop") as mock_stop:
                dev.samx.setpoint.set(1)
        mock_rpc.assert_called_once()
        mock_stop.assert_called_once()


@pytest.fixture
def device_config():
    return {
        "id": "1c6518b2-b779-4b28-b8b1-31295f8fbf26",
        "accessGroups": "customer",
        "name": "eiger",
        "sessionId": "569ea788-09d7-44fc-a140-b0b34a2b7f6f",
        "enabled": True,
        "readOnly": False,
        "readoutPriority": "monitored",
        "deviceClass": "SimCamera",
        "deviceConfig": {"device_access": True, "labels": "eiger", "name": "eiger"},
        "deviceTags": ["detector"],
    }


@pytest.fixture
def device_obj(device_config):
    service_mock = mock.MagicMock()
    service_mock.connector = ConnectorMock("")
    dm = DeviceManagerBase(service_mock)
    info = get_device_info_mock(device_config["name"], device_config["deviceClass"])
    dm._add_device(device_config, info)
    obj = dm.devices[device_config["name"]]
    yield obj


def test_create_device_saves_config(device_obj, device_config):
    obj = device_obj
    assert obj._config == device_config


def test_device_enabled(device_obj, device_config):
    obj = device_obj
    assert obj.enabled == device_config["enabled"]
    device_config["enabled"] = False
    assert obj.enabled == device_config["enabled"]


def test_device_enable(device_obj):
    obj = device_obj
    with mock.patch.object(obj.parent.config_helper, "send_config_request") as config_req:
        obj.enabled = True
        config_req.assert_called_once_with(action="update", config={obj.name: {"enabled": True}})


def test_device_enable_set(device_obj):
    obj = device_obj
    with mock.patch.object(obj.parent.config_helper, "send_config_request") as config_req:
        obj.read_only = False
        config_req.assert_called_once_with(action="update", config={obj.name: {"readOnly": False}})


@pytest.mark.parametrize(
    "val,raised_error",
    [({"in": 5}, None), ({"in": 5, "out": 10}, None), ({"5", "4"}, TypeCheckError)],
)
def test_device_set_user_parameter(device_obj, val, raised_error):
    obj = device_obj
    with mock.patch.object(obj.parent.config_helper, "send_config_request") as config_req:
        if raised_error is None:
            obj.set_user_parameter(val)
            config_req.assert_called_once_with(
                action="update", config={obj.name: {"userParameter": val}}
            )
        else:
            with pytest.raises(raised_error):
                obj.set_user_parameter(val)


@pytest.mark.parametrize(
    "user_param,val,out,raised_error",
    [
        ({"in": 2, "out": 5}, {"in": 5}, {"in": 5, "out": 5}, None),
        ({"in": 2, "out": 5}, {"in": 5, "out": 10}, {"in": 5, "out": 10}, None),
        ({"in": 2, "out": 5}, {"5", "4"}, None, TypeCheckError),
        (None, {"in": 5}, {"in": 5}, None),
    ],
)
def test_device_update_user_parameter(device_obj, user_param, val, out, raised_error):
    obj = device_obj
    obj._config["userParameter"] = user_param
    with mock.patch.object(obj.parent.config_helper, "send_config_request") as config_req:
        if raised_error is None:
            obj.update_user_parameter(val)
            config_req.assert_called_once_with(
                action="update", config={obj.name: {"userParameter": out}}
            )
        else:
            with pytest.raises(raised_error):
                obj.update_user_parameter(val)


def test_status_wait():
    connector = mock.MagicMock()

    def lrange_mock(*args, **kwargs):
        yield False
        yield True

    def get_lrange(*args, **kwargs):
        return next(lmock)

    lmock = lrange_mock()
    connector.lrange = get_lrange
    status = Status(connector, "test")
    status.wait()


def test_status_wait_raises_timeout():
    connector = mock.MagicMock()
    connector.lrange.return_value = False
    status = Status(connector, "test")
    with pytest.raises(TimeoutError):
        status.wait(timeout=0.1)


def test_device_get_device_config():
    device = DeviceBase(name="test", config={"deviceConfig": {"tolerance": 1}})
    assert device.get_device_config() == {"tolerance": 1}


def test_device_set_device_config():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"deviceConfig": {"tolerance": 1}}, parent=parent)
    device.set_device_config({"tolerance": 2})
    assert device.get_device_config() == {"tolerance": 2}
    parent.config_helper.send_config_request.assert_called_once()


def test_get_device_tags():
    device = DeviceBase(name="test", config={"deviceTags": ["tag1", "tag2"]})
    assert device.get_device_tags() == ["tag1", "tag2"]

    device = DeviceBase(name="test", config={})
    assert device.get_device_tags() == []


def test_set_device_tags():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"deviceTags": ["tag1", "tag2"]}, parent=parent)
    device.set_device_tags(["tag3", "tag4"])
    assert device.get_device_tags() == ["tag3", "tag4"]
    parent.config_helper.send_config_request.assert_called_once()


def test_add_device_tag():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"deviceTags": ["tag1", "tag2"]}, parent=parent)
    device.add_device_tag("tag3")
    assert device.get_device_tags() == ["tag1", "tag2", "tag3"]
    parent.config_helper.send_config_request.assert_called_once()


def test_add_device_tags_duplicate():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"deviceTags": ["tag1", "tag2"]}, parent=parent)
    device.add_device_tag("tag1")
    assert device.get_device_tags() == ["tag1", "tag2"]
    parent.config_helper.send_config_request.assert_not_called()


def test_remove_device_tag():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"deviceTags": ["tag1", "tag2"]}, parent=parent)
    device.remove_device_tag("tag1")
    assert device.get_device_tags() == ["tag2"]
    parent.config_helper.send_config_request.assert_called_once()


def test_device_wm():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"deviceTags": ["tag1", "tag2"]}, parent=parent)
    with mock.patch.object(parent.devices, "wm", new_callable=mock.PropertyMock) as wm:
        res = device.wm
        parent.devices.wm.assert_called_once()


def test_readout_priority():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"readoutPriority": "baseline"}, parent=parent)
    assert device.readout_priority == "baseline"


def test_set_readout_priority():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"readoutPriority": "baseline"}, parent=parent)
    device.readout_priority = "monitored"
    assert device.readout_priority == "monitored"
    parent.config_helper.send_config_request.assert_called_once()


def test_on_failure():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"onFailure": "buffer"}, parent=parent)
    assert device.on_failure == "buffer"


def test_set_on_failure():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"onFailure": "buffer"}, parent=parent)
    device.on_failure = "retry"
    assert device.on_failure == "retry"
    parent.config_helper.send_config_request.assert_called_once()


def test_read_only():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"read_only": False}, parent=parent)
    assert device.read_only is False


def test_set_read_only():
    parent = mock.MagicMock()
    device = DeviceBase(name="test", config={"read_only": False}, parent=parent)
    device.read_only = True
    assert device.read_only is True
    parent.config_helper.send_config_request.assert_called_once()


def test_device_container_wm():
    devs = DeviceContainer()
    devs["test"] = Device(name="test", config={}, parent=mock.MagicMock())
    with mock.patch.object(devs.test, "read", return_value={"test": {"value": 1}}) as read:
        devs.wm("test")
        devs.wm("tes*")


def test_device_container_wm_with_setpoint():
    devs = DeviceContainer()
    devs["test"] = Device(name="test", config={}, parent=mock.MagicMock())
    with mock.patch.object(
        devs.test, "read", return_value={"test": {"value": 1}, "test_setpoint": {"value": 1}}
    ) as read:
        devs.wm("test")


def test_device_container_wm_with_user_setpoint():
    devs = DeviceContainer()
    devs["test"] = Device(name="test", config={}, parent=mock.MagicMock())
    with mock.patch.object(
        devs.test, "read", return_value={"test": {"value": 1}, "test_user_setpoint": {"value": 1}}
    ) as read:
        devs.wm("test")


@pytest.mark.parametrize("device_cls", [Device, Signal, Positioner])
def test_device_has_describe_method(device_cls):
    devs = DeviceContainer()
    parent = mock.MagicMock(spec=DeviceManagerBase)
    devs["test"] = device_cls(name="test", config={}, parent=parent)
    assert hasattr(devs.test, "describe")
    with mock.patch.object(devs.test, "_run_rpc_call") as mock_rpc:
        devs.test.describe()
        mock_rpc.assert_not_called()


@pytest.mark.parametrize("device_cls", [Device, Signal, Positioner])
def test_device_has_describe_configuration_method(device_cls):
    devs = DeviceContainer()
    parent = mock.MagicMock(spec=DeviceManagerBase)
    devs["test"] = device_cls(name="test", config={}, parent=parent)
    assert hasattr(devs.test, "describe_configuration")
    with mock.patch.object(devs.test, "_run_rpc_call") as mock_rpc:
        devs.test.describe_configuration()
        mock_rpc.assert_not_called()


def test_show_all():
    # Create a mock Console object
    console = mock.MagicMock()
    parent = mock.MagicMock()
    parent.parent = mock.MagicMock(spec=DeviceManagerBase)

    # Create a DeviceContainer with some mock Devices
    devs = DeviceContainer()
    devs["dev1"] = DeviceBase(
        name="dev1",
        config={
            "description": "Device 1",
            "enabled": True,
            "readOnly": False,
            "deviceClass": "Class1",
            "readoutPriority": "high",
            "deviceTags": ["tag1", "tag2"],
        },
        parent=parent,
    )
    devs["dev2"] = DeviceBase(
        name="dev2",
        config={
            "description": "Device 2",
            "enabled": False,
            "readOnly": True,
            "deviceClass": "Class2",
            "readoutPriority": "low",
            "deviceTags": ["tag3", "tag4"],
        },
        parent=parent,
    )

    # Call show_all with the mock Console
    devs.show_all(console)

    # check that the device names were printed
    table = console.print.call_args[0][0]
    assert len(table.rows) == 2
    assert list(table.columns[0].cells) == ["dev1", "dev2"]
    # Check that Console.print was called with a Table containing the correct data
    console.print.assert_called_once()


def test_adjustable_mixin_limits():
    adj = AdjustableMixin()
    adj.root = mock.MagicMock()
    adj.root.parent.connector.get.return_value = messages.DeviceMessage(
        signals={"low": {"value": -12}, "high": {"value": 12}}, metadata={}
    )
    assert adj.limits == [-12, 12]


def test_adjustable_mixin_limits_missing():
    adj = AdjustableMixin()
    adj.root = mock.MagicMock()
    adj.root.parent.connector.get.return_value = None
    assert adj.limits == [0, 0]


def test_adjustable_mixin_set_limits():
    adj = AdjustableMixin()
    adj.update_config = mock.MagicMock()
    adj.limits = [-12, 12]
    adj.update_config.assert_called_once_with({"deviceConfig": {"limits": [-12, 12]}})


def test_adjustable_mixin_set_low_limit():
    adj = AdjustableMixin()
    adj.update_config = mock.MagicMock()
    adj.root = mock.MagicMock()
    adj.root.parent.connector.get.return_value = messages.DeviceMessage(
        signals={"low": {"value": -12}, "high": {"value": 12}}, metadata={}
    )
    adj.low_limit = -20
    adj.update_config.assert_called_once_with({"deviceConfig": {"limits": [-20, 12]}})


def test_adjustable_mixin_set_high_limit():
    adj = AdjustableMixin()
    adj.update_config = mock.MagicMock()
    adj.root = mock.MagicMock()
    adj.root.parent.connector.get.return_value = messages.DeviceMessage(
        signals={"low": {"value": -12}, "high": {"value": 12}}, metadata={}
    )
    adj.high_limit = 20
    adj.update_config.assert_called_once_with({"deviceConfig": {"limits": [-12, 20]}})


def test_computed_signal_set_compute_method():
    comp_signal = ComputedSignal(name="comp_signal", parent=mock.MagicMock())

    def my_compute_method():
        return "a + b"

    with mock.patch.object(comp_signal, "update_config") as update_config:
        comp_signal.set_compute_method(my_compute_method)
        update_config.assert_called_once_with(
            {
                "deviceConfig": {
                    "compute_method": '    def my_compute_method():\n        return "a + b"\n'
                }
            }
        )


def test_computed_signal_set_signals():
    comp_signal = ComputedSignal(name="comp_signal", parent=mock.MagicMock())
    with mock.patch.object(comp_signal, "update_config") as update_config:
        comp_signal.set_input_signals(
            Signal(name="a", parent=mock.MagicMock(spec=DeviceManagerBase)),
            Signal(name="b", parent=mock.MagicMock(spec=DeviceManagerBase)),
        )
        update_config.assert_called_once_with({"deviceConfig": {"input_signals": ["a", "b"]}})


def test_computed_signal_set_signals_raises_error():
    comp_signal = ComputedSignal(name="comp_signal", parent=mock.MagicMock())
    with pytest.raises(ValueError):
        comp_signal.set_input_signals("a", "b")


def test_computed_signal_set_signals_empty():
    comp_signal = ComputedSignal(name="comp_signal", parent=mock.MagicMock())
    with mock.patch.object(comp_signal, "update_config") as update_config:
        comp_signal.set_input_signals()
        update_config.assert_called_once_with({"deviceConfig": {"input_signals": []}})


def test_computed_signal_raises_error_on_set_compute_method():
    comp_signal = ComputedSignal(name="comp_signal", parent=mock.MagicMock())
    with pytest.raises(ValueError):
        comp_signal.set_compute_method("a + b")



"""
This module provides a helper class for updating and saving the BEC device configuration.
"""

from __future__ import annotations

import datetime
import json
import os
import pathlib
import time
import uuid
from typing import TYPE_CHECKING

import yaml

import bec_lib
from bec_lib import messages
from bec_lib.bec_errors import DeviceConfigError, ServiceConfigError
from bec_lib.bec_yaml_loader import yaml_load
from bec_lib.endpoints import MessageEndpoints
from bec_lib.file_utils import DeviceConfigWriter
from bec_lib.logger import bec_logger
from bec_lib.messages import DeviceConfigMessage, RequestResponseMessage

if TYPE_CHECKING:
    from bec_lib.redis_connector import RedisConnector

logger = bec_logger.logger


class ConfigHelper:
    """Config Helper"""

    def __init__(self, connector: RedisConnector, service_name: str = None) -> None:
        """Helper class for updating and saving the BEC device configuration.

        Args:
            connector (RedisConnector): Redis connector.
            service_name (str, optional): Name of the service. Defaults to None.
        """
        self.connector = connector
        self._service_name = service_name
        self.writer_mixin = None
        self._base_path_recovery = None

    def update_session_with_file(self, file_path: str, save_recovery: bool = True) -> None:
        """Update the current session with a yaml file from disk.

        Args:
            file_path (str): Full path to the yaml file.
            save_recovery (bool, optional): Save the current session before updating. Defaults to True.
        """
        if save_recovery:
            time_stamp = f"{datetime.datetime.now():%Y-%m-%d_%H-%M-%S}"
            if not self._base_path_recovery:
                self._update_base_path_recovery()
            if os.path.exists(self._base_path_recovery) is False:
                self.writer_mixin.create_directory(self._base_path_recovery)
            fname = os.path.join(self._base_path_recovery, f"recovery_config_{time_stamp}.yaml")
            success = self._save_config_to_file(fname, raise_on_error=False)
            if success:
                print(f"A recovery config was written to {fname}.")
        config = self._load_config_from_file(file_path)
        self.send_config_request(action="set", config=config)

    def _update_base_path_recovery(self):
        """
        Compile the filepath for the recovery configs.
        """
        # pylint: disable=import-outside-toplevel
        from bec_lib.bec_service import SERVICE_CONFIG

        service_cfg = SERVICE_CONFIG.config["service_config"].get("log_writer", None)
        if not service_cfg:
            raise ServiceConfigError(
                f"ServiceConfig {service_cfg} must at least contain key with 'log_writer'"
            )
        self.writer_mixin = DeviceConfigWriter(service_cfg)
        self._base_path_recovery = self.writer_mixin.get_recovery_directory()
        self.writer_mixin.create_directory(self._base_path_recovery)

    def _load_config_from_file(self, file_path: str) -> dict:
        data = {}
        if pathlib.Path(file_path).suffix not in (".yaml", ".yml"):
            raise NotImplementedError

        with open(file_path, "r", encoding="utf-8") as stream:
            try:
                data = yaml_load(stream)
                logger.trace(
                    f"Loaded new config from disk: {json.dumps(data, sort_keys=True, indent=4)}"
                )
            except yaml.YAMLError as err:
                logger.error(f"Error while loading config from disk: {repr(err)}")

        return data

    def save_current_session(self, file_path: str):
        """Save the current session as a yaml file to disk.

        Args:
            file_path (str): Full path to the yaml file.
        """
        self._save_config_to_file(file_path)
        print(f"Config was written to {file_path}.")

    def _save_config_to_file(self, file_path: str, raise_on_error: bool = True) -> bool:
        config = self.connector.get(MessageEndpoints.device_config())
        if not config:
            if raise_on_error:
                raise DeviceConfigError("No config found in the session.")
            return False
        config = config.content["resource"]
        out = {}
        for dev in config:
            dev.pop("id", None)
            dev.pop("createdAt", None)
            dev.pop("createdBy", None)
            dev.pop("sessionId", None)
            name = dev.pop("name")
            out[name] = dev

        with open(file_path, "w") as file:
            file.write(yaml.dump(out))
        return True

    def send_config_request(self, action: str = "update", config=None) -> None:
        """
        send request to update config
        Returns:

        """
        if action in ["update", "add", "set"] and not config:
            raise DeviceConfigError(f"Config cannot be empty for an {action} request.")
        RID = str(uuid.uuid4())
        self.connector.send(
            MessageEndpoints.device_config_request(),
            DeviceConfigMessage(action=action, config=config, metadata={"RID": RID}),
        )

        reply = self.wait_for_config_reply(RID)

        if not reply.content["accepted"] and not reply.metadata.get("updated_config"):
            raise DeviceConfigError(
                f"Failed to update the config: {reply.content['message']}. No devices were updated."
            )
        try:
            if not reply.content["accepted"] and reply.metadata.get("updated_config"):
                raise DeviceConfigError(
                    f"Failed to update the config: {reply.content['message']}. The old config will be kept in the device config history."
                )

            if "failed_devices" in reply.metadata:
                print("Failed to update the config for some devices.")
                for dev in reply.metadata["failed_devices"]:
                    print(
                        f"Device {dev} failed to update:\n {reply.metadata['failed_devices'][dev]}."
                    )
                devices = [dev for dev in reply.metadata["failed_devices"]]

                raise DeviceConfigError(
                    f"Failed to update the config for some devices. The following devices were disabled: {devices}."
                )
        finally:
            # wait for the device server and scan server to acknowledge the config change
            self.wait_for_service_response(RID)

    def wait_for_service_response(self, RID: str, timeout=10) -> messages.ServiceResponseMessage:
        """
        wait for service response

        Args:
            RID (str): request id
            timeout (int, optional): timeout in seconds. Defaults to 10.

        Returns:
            ServiceResponseMessage: reply message
        """
        elapsed_time = 0
        max_time = timeout
        while True:
            service_messages = self.connector.lrange(MessageEndpoints.service_response(RID), 0, -1)
            if not service_messages:
                time.sleep(0.005)
                elapsed_time += 0.005
            else:
                ack_services = [
                    msg.content["response"]["service"]
                    for msg in service_messages
                    if msg is not None
                ]
                checked_services = set(["DeviceServer", "ScanServer"])
                if self._service_name:
                    checked_services.add(self._service_name)
                if checked_services.issubset(set(ack_services)):
                    break
            if elapsed_time > max_time:
                if service_messages:
                    raise DeviceConfigError(
                        "Timeout reached whilst waiting for config change to be acknowledged."
                        f" Received {service_messages}."
                    )

                raise DeviceConfigError(
                    "Timeout reached whilst waiting for config change to be acknowledged. No"
                    " messages received."
                )

    def wait_for_config_reply(self, RID: str, timeout=10) -> RequestResponseMessage:
        """
        wait for config reply

        Args:
            RID (str): request id
            timeout (int, optional): timeout in seconds. Defaults to 10.

        Returns:
            RequestResponseMessage: reply message
        """
        start = 0
        while True:
            msg = self.connector.get(MessageEndpoints.device_config_request_response(RID))
            if msg is None:
                time.sleep(0.01)
                start += 0.01

                if start > timeout:
                    raise DeviceConfigError("Timeout reached whilst waiting for config reply.")
                continue
            return msg

    def load_demo_config(self):
        """Load BEC device demo_config.yaml for simulation."""
        dir_path = os.path.abspath(os.path.join(os.path.dirname(bec_lib.__file__), "./configs/"))
        fpath = os.path.join(dir_path, "demo_config.yaml")
        self.update_session_with_file(fpath)



"""
This module contains functions to serialize and deserialize function signatures.
"""

from __future__ import annotations

import ast
import builtins
import inspect
import operator
import sys
from collections.abc import Callable
from typing import Any, Literal, Union

from bec_lib.device import DeviceBase
from bec_lib.scan_items import ScanItem


def serialize_dtype(dtype: type) -> Any:
    """
    Convert a dtype to a string.

    Args:
        dtype (type): Data type

    Returns:
        str: String representation of the data type
    """
    if hasattr(dtype, "__name__"):
        name = dtype.__name__
        # changed in python 3.10. Refactor this when we upgrade
        if name not in ["Literal", "Union"]:
            return name
    if hasattr(dtype, "__module__"):
        if dtype.__module__ == "typing":
            if dtype.__class__.__name__ == "_UnionGenericAlias":
                return " | ".join([serialize_dtype(x) for x in dtype.__args__])
            if dtype.__class__.__name__ == "_LiteralGenericAlias":
                return {"Literal": dtype.__args__}
    if isinstance(dtype, str):
        if sys.version_info[:3] >= (3, 10):
            return dtype
        # remove this when we upgrade to python 3.10
        ####
        if dtype.startswith("typing.Literal["):
            return {"Literal": ast.literal_eval(dtype[15:-1])}
        if dtype.startswith("Literal["):
            return {"Literal": ast.literal_eval(dtype[8:-1])}
        return dtype
        ####
    raise ValueError(f"Unknown dtype {dtype}")


def deserialize_dtype(dtype: Any) -> type:
    """
    Convert a serialized dtype to a type.

    Args:
        dtype (str): String representation of the data type

    Returns:
        type: Data type
    """
    if dtype == "_empty":
        # pylint: disable=protected-access
        return inspect._empty
    if isinstance(dtype, dict):
        if "Literal" in dtype:
            # remove this when we upgrade to python 3.10

            #### remove this section
            literal = Literal[str(dtype)]
            literal.__args__ = dtype["Literal"]
            return literal
            #### remove this section

            #### add this section when we upgrade to python 3.10
            # return Literal[*dtype["Literal"]]
            #### add this section when we upgrade to python 3.10

        raise ValueError(f"Unknown dtype {dtype}")
    if isinstance(dtype, str) and "|" in dtype:
        # remove this when we upgrade to python 3.10
        if sys.version_info[:3] < (3, 10):
            union = Union[int, str]
            union.__args__ = [deserialize_dtype(x.strip()) for x in dtype.split("|")]
            return union
        return operator.or_(*[deserialize_dtype(x.strip()) for x in dtype.split("|")])

    builtin_type = builtins.__dict__.get(dtype)
    if builtin_type:
        return builtin_type
    if dtype == "DeviceBase":
        return DeviceBase
    if dtype == "ScanItem":
        return ScanItem


def signature_to_dict(func: Callable, include_class_obj=False) -> dict:
    """
    Convert a function signature to a dictionary.
    The dictionary can be used to reconstruct the signature using dict_to_signature.

    Args:
        func (Callable): Function to be converted

    Returns:
        dict: Dictionary representation of the function signature
    """
    out = []
    params = inspect.signature(func).parameters
    for param_name, param in params.items():
        if not include_class_obj and param_name == "self" or param_name == "cls":
            continue
        # pylint: disable=protected-access
        out.append(
            {
                "name": param_name,
                "kind": param.kind.name,
                "default": param.default if param.default != inspect._empty else "_empty",
                "annotation": serialize_dtype(param.annotation),
            }
        )
    return out


def dict_to_signature(params: list[dict]) -> inspect.Signature:
    """
    Convert a dictionary representation of a function signature to a signature object.

    Args:
        params (list[dict]): List of dictionaries representing the function signature

    Returns:
        inspect.Signature: Signature object
    """
    out = []
    for param in params:
        # pylint: disable=protected-access
        out.append(
            inspect.Parameter(
                name=param["name"],
                kind=getattr(inspect.Parameter, param["kind"]),
                default=param["default"] if param["default"] != "_empty" else inspect._empty,
                annotation=deserialize_dtype(param["annotation"]),
            )
        )
    return inspect.Signature(out)



"""
This module provides the Observer class and the ObserverManager class. The Observer class is used
to define an observer object, which is used to monitor a device for a specific condition. The
ObserverManager class is used to manage the observer objects.
"""

from __future__ import annotations

import enum
from typing import TYPE_CHECKING

from typeguard import typechecked

from bec_lib.device import Device
from bec_lib.endpoints import MessageEndpoints

if TYPE_CHECKING:
    from bec_lib.devicemanager import DeviceManagerBase


class ObserverAction(str, enum.Enum):
    PAUSE = "pause"
    RESUME = "resume"
    RESTART = "restart"


class Observer:
    def __init__(
        self,
        name: str,
        device: str,
        on_trigger: ObserverAction = None,
        on_resume: ObserverAction = None,
        limits: list = None,
        low_limit: float = None,
        high_limit: float = None,
        target_value=None,
    ):
        self.name = name
        self.device = device
        self.on_trigger = on_trigger
        self.on_resume = on_resume
        self._limits = [None, None]

        self._check_limits(limits, low_limit, high_limit)

        self.target_value = target_value
        self._enabled = False

        self._check_device()
        self._check_trigger()
        self._check_targets()

    @property
    def limits(self):
        return self._limits

    @limits.setter
    @typechecked
    def limits(self, val: list):
        self._limits = val

    @property
    def low_limit(self):
        return self.limits[0]

    @low_limit.setter
    @typechecked
    def low_limit(self, val: float):
        self.limits[0] = val

    @property
    def high_limit(self):
        return self.limits[1]

    @high_limit.setter
    @typechecked
    def high_limit(self, val: float):
        self.limits[1] = val

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        self._enabled = val

    def _check_limits(self, limits, low_limit, high_limit):
        if limits is not None and (low_limit is not None or high_limit is not None):
            raise AttributeError(
                "Ambiguous condition: Limits are set multiple times. Use either limits or"
                " low_limit/high_limit."
            )
        if limits is not None:
            self.limits = limits
        if low_limit is not None:
            self.low_limit = low_limit
        if high_limit is not None:
            self.high_limit = high_limit

    def _check_device(self):
        if isinstance(self.device, Device):
            self.device = self.device.name

    def _check_trigger(self):
        if not isinstance(self.on_trigger, ObserverAction):
            self.on_trigger = ObserverAction(self.on_trigger)
        if not isinstance(self.on_resume, ObserverAction):
            self.on_resume = ObserverAction(self.on_resume)

    def _check_targets(self):
        if self.limits is not None and self.target_value is not None:
            raise AttributeError(
                "Ambiguous condition: Only one target (limits, target_value) can be set."
            )
        if all(val is None for val in self.limits) and self.target_value is None:
            raise AttributeError("No condition set.")

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "device": self.device,
            "on_trigger": self.on_trigger,
            "on_resume": self.on_resume,
            "limits": self.limits,
            "target_value": self.target_value,
        }

    @classmethod
    def from_dict(cls, config: dict) -> Observer:
        return cls(**config)


class ObserverManager:
    def __init__(self, device_manager: DeviceManagerBase):
        self.device_manager = device_manager
        self._observer = self._get_installed_observer()

    @typechecked
    def add_observer(self, observer: Observer, ignore_existing: bool = False):
        if not hasattr(self.device_manager.devices, observer.device):
            AttributeError(
                f"The specified observer uses device {observer.device} which is currently not"
                " configured."
            )
        if self._is_device_observed(observer.device) and not ignore_existing:
            raise AttributeError(
                f"Device {observer.device} is already being observed. If you want to add an"
                " additional observer for this device, use 'ignore_existing=True'."
            )
        self._observer.append(observer)
        self.update_observer()

    def _is_device_observed(self, device: str) -> bool:
        return any(obs.device == device for obs in self._observer)

    def _get_installed_observer(self):
        # get current observer list from Redis
        observer_msg = self.device_manager.connector.get(MessageEndpoints.observer())
        if observer_msg is None:
            return []
        return [Observer.from_dict(obs) for obs in observer_msg.content["observer"]]

    def update_observer(self):
        # send the current observer list to MongoDB and Redis
        pass

    def list_observer(self):
        pass



import traceback
from contextlib import redirect_stdout
from io import StringIO
from typing import Any

import ophyd

from bec_lib import Alarms, MessageEndpoints, bec_logger, messages
from bec_server.device_server.devices import is_serializable, rgetattr

logger = bec_logger.logger


class RPCMixin:
    """Mixin for handling RPC calls"""

    def run_rpc(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Run RPC call and send result to client. RPC calls also capture stdout and
        stderr and send it to the client.

        Args:
            instr: DeviceInstructionMessage

        """
        result = StringIO()
        with redirect_stdout(result):
            try:
                instr_params = instr.content.get("parameter")
                device = instr.content["device"]
                self._assert_device_is_enabled(instr)
                res = self._process_rpc_instruction(instr)
                # send result to client
                self._send_rpc_result_to_client(device, instr_params, res, result)
                logger.trace(res)
            except Exception as exc:  # pylint: disable=broad-except
                # send error to client
                self._send_rpc_exception(exc, instr)

    def _get_result_from_rpc(self, rpc_var: Any, instr_params: dict) -> Any:
        if callable(rpc_var):
            args = tuple(instr_params.get("args", ()))
            kwargs = instr_params.get("kwargs", {})
            if args and kwargs:
                res = rpc_var(*args, **kwargs)
            elif args:
                res = rpc_var(*args)
            elif kwargs:
                res = rpc_var(**kwargs)
            else:
                res = rpc_var()
        else:
            res = rpc_var
        if not is_serializable(res):
            if isinstance(res, ophyd.StatusBase):
                return res
            if isinstance(res, list) and instr_params.get("func") in ["stage", "unstage"]:
                # pylint: disable=protected-access
                return [obj._staged for obj in res]
            res = None
            self.connector.raise_alarm(
                severity=Alarms.WARNING,
                alarm_type="TypeError",
                source=instr_params,
                msg=f"Return value of rpc call {instr_params} is not serializable.",
                metadata={},
            )
        return res

    def _send_rpc_result_to_client(
        self, device: str, instr_params: dict, res: Any, result: StringIO
    ):
        self.connector.set(
            MessageEndpoints.device_rpc(instr_params.get("rpc_id")),
            messages.DeviceRPCMessage(
                device=device, return_val=res, out=result.getvalue(), success=True
            ),
            expire=1800,
        )

    def _rpc_read_and_return(self, instr: messages.DeviceInstructionMessage) -> Any:
        res = self._read_and_update_devices([instr.content["device"]], instr.metadata)
        if isinstance(res, list) and len(res) == 1:
            res = res[0]
        return res

    def _rpc_read_configuration_and_return(self, instr: messages.DeviceInstructionMessage) -> Any:
        res = self._read_config_and_update_devices([instr.content["device"]], instr.metadata)
        if isinstance(res, list) and len(res) == 1:
            res = res[0]
        return res

    def _process_rpc_instruction(self, instr: messages.DeviceInstructionMessage) -> Any:
        # handle ophyd read. This is a special case because we also want to update the
        # buffered value in redis
        instr_params = instr.content.get("parameter")
        device_root = instr.content["device"].split(".")[0]
        if instr_params.get("func") == "read" or instr_params.get("func").endswith(".read"):
            if instr_params.get("func") == "read":
                obj = self.device_manager.devices[device_root].obj
            else:
                obj = rgetattr(
                    self.device_manager.devices[device_root].obj,
                    instr_params.get("func").split(".read")[0],
                )
            if isinstance(obj, ophyd.Device):
                return self._rpc_read_and_return(instr)
            if isinstance(obj, ophyd.Signal):
                if obj.kind not in [ophyd.Kind.omitted, ophyd.Kind.config]:
                    return self._rpc_read_and_return(instr)
                if obj.kind == ophyd.Kind.config:
                    return self._rpc_read_configuration_and_return(instr)
                if obj.kind == ophyd.Kind.omitted:
                    return obj.read()
        if instr_params.get("func") == "read_configuration" or instr_params.get("func").endswith(
            ".read_configuration"
        ):
            return self._rpc_read_configuration_and_return(instr)
        if instr_params.get("kwargs", {}).get("_set_property"):
            sub_access = instr_params.get("func").split(".")
            property_name = sub_access[-1]
            if len(sub_access) > 1:
                sub_access = sub_access[0:-1]
            else:
                sub_access = []
            obj = self.device_manager.devices[device_root].obj
            if sub_access:
                obj = rgetattr(obj, ".".join(sub_access))
            setattr(obj, property_name, instr_params.get("args")[0])
            return None

        # handle other ophyd methods
        rpc_var = rgetattr(self.device_manager.devices[device_root].obj, instr_params.get("func"))
        res = self._get_result_from_rpc(rpc_var, instr_params)
        if instr_params.get("func") == "put" or instr_params.get("func").endswith(".put"):
            if instr_params.get("func") == "put":
                obj = self.device_manager.devices[device_root].obj
            else:
                obj = rgetattr(
                    self.device_manager.devices[device_root].obj,
                    instr_params.get("func").split(".put")[0],
                )
            if obj.kind == ophyd.Kind.config:
                return self._rpc_read_configuration_and_return(instr)
            if obj.kind in [ophyd.Kind.normal, ophyd.Kind.hinted]:
                return self._rpc_read_and_return(instr)
        if isinstance(res, ophyd.StatusBase):
            res.__dict__["instruction"] = instr
            res.add_callback(self._status_callback)
            res = {
                "type": "status",
                "RID": instr.metadata.get("RID"),
                "success": res.success,
                "timeout": res.timeout,
                "done": res.done,
                "settle_time": res.settle_time,
            }
        elif isinstance(res, tuple) and hasattr(res, "_asdict") and hasattr(res, "_fields"):
            # convert namedtuple to dict
            res = {
                "type": "namedtuple",
                "RID": instr.metadata.get("RID"),
                "fields": res._fields,
                "values": res._asdict(),
            }
        elif isinstance(res, list) and res and isinstance(res[0], ophyd.Staged):
            res = [str(stage) for stage in res]
        return res

    def _send_rpc_exception(self, exc: Exception, instr: messages.DeviceInstructionMessage):
        exc_formatted = {
            "error": exc.__class__.__name__,
            "msg": exc.args,
            "traceback": traceback.format_exc(),
        }
        logger.info(f"Received exception: {exc_formatted}, {exc}")
        instr_params = instr.content.get("parameter")
        self.connector.set(
            MessageEndpoints.device_rpc(instr_params.get("rpc_id")),
            messages.DeviceRPCMessage(
                device=instr.content["device"], return_val=None, out=exc_formatted, success=False
            ),
        )



"""
This module defines the interface for a connector
"""

from __future__ import annotations

import abc

from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.messages import BECMessage, LogMessage

logger = bec_logger.logger


class ConsumerConnectorError(Exception):
    """
    ConsumerConnectorError is raised when there is an error with the connector
    """


class MessageObject:
    """
    MessageObject is a wrapper for a message and its topic
    """

    def __init__(self, topic: str, value: BECMessage) -> None:
        self.topic = topic
        self._value = value

    @property
    def value(self) -> BECMessage:
        """
        Get the message
        """
        return self._value

    def __eq__(self, ref_val: MessageObject) -> bool:
        if not isinstance(ref_val, MessageObject):
            return False
        return self._value == ref_val.value and self.topic == ref_val.topic

    def __str__(self):
        return f"MessageObject(topic={self.topic}, value={self._value})"


class StoreInterface(abc.ABC):
    """StoreBase defines the interface for storing data"""

    def __init__(self, store):
        pass

    @abc.abstractmethod
    def pipeline(self):
        """Create a pipeline for batch operations"""

    @abc.abstractmethod
    def execute_pipeline(self, pipeline):
        """Execute a pipeline"""

    @abc.abstractmethod
    def lpush(
        self, topic: str, msg: str, pipe=None, max_size: int = None, expire: int = None
    ) -> None:
        """Push a message to the left of the list"""

    @abc.abstractmethod
    def lset(self, topic: str, index: int, msg: str, pipe=None) -> None:
        """Set a value in the list at the given index"""

    @abc.abstractmethod
    def rpush(self, topic: str, msg: str, pipe=None) -> int:
        """Push a message to the right of the list"""

    @abc.abstractmethod
    def lrange(self, topic: str, start: int, end: int, pipe=None):
        """Get a range of values from the list"""

    @abc.abstractmethod
    def set(self, topic: str, msg, pipe=None, expire: int = None) -> None:
        """Set a value"""

    @abc.abstractmethod
    def keys(self, pattern: str) -> list:
        """Get keys that match the pattern"""

    @abc.abstractmethod
    def delete(self, topic, pipe=None):
        """Delete a key"""

    @abc.abstractmethod
    def get(self, topic: str, pipe=None):
        """Get a value"""

    @abc.abstractmethod
    def xadd(self, topic: str, msg_dict: dict, max_size=None, pipe=None, expire: int = None):
        """Add a message to the stream"""

    @abc.abstractmethod
    def xread(
        self,
        topic: str,
        id: str = None,
        count: int = None,
        block: int = None,
        pipe=None,
        from_start=False,
    ) -> list:
        """Read from the stream"""

    @abc.abstractmethod
    def xrange(self, topic: str, min: str, max: str, count: int = None, pipe=None):
        """Read from the stream"""


class PubSubInterface(abc.ABC):
    """PubSubBase defines the interface for a pub/sub connector"""

    @abc.abstractmethod
    def raw_send(self, topic: str, msg: bytes) -> None:
        """Send a raw message without using the BECMessage class"""

    @abc.abstractmethod
    def send(self, topic: str, msg: BECMessage) -> None:
        """Send a message"""

    @abc.abstractmethod
    def register(self, topics=None, patterns=None, cb=None, start_thread=True, **kwargs):
        """Register a callback for a topic or pattern"""

    @abc.abstractmethod
    def unregister(self, topics=None, pattern=None, cb=None):
        """Unregister a callback for a topic or pattern"""

    @abc.abstractmethod
    def poll_messages(self, timeout=None):
        """Poll for new messages, receive them and execute callbacks"""


class ConnectorBase(PubSubInterface, StoreInterface):
    """ConnectorBase defines the interface for a connector"""

    def raise_warning(self, msg):
        """Raise a warning"""
        raise NotImplementedError

    def send_client_info(self, msg):
        """send a msg to the client, will automatically be logged too."""
        raise NotImplementedError

    def set_and_publish(self, topic: str, msg, pipe=None, expire: int = None) -> None:
        """Set a value and publish it"""
        raise NotImplementedError

    def shutdown(self):
        """Shutdown the connector"""



import pytest

from bec_lib.logger import bec_logger
from bec_lib.tests.fixtures import dm_with_devices
from bec_server.scan_server.tests.utils import ScanServerMock


# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
@pytest.fixture
def scan_server_mock(dm_with_devices):
    server = ScanServerMock(dm_with_devices)
    yield server
    bec_logger.logger.remove()



from __future__ import annotations

import collections
import time
from typing import TYPE_CHECKING

from bec_ipython_client.callbacks.scan_progress import LiveUpdatesScanProgress
from bec_lib import bec_logger
from bec_lib.bec_errors import ScanInterruption

from .live_table import LiveUpdatesTable
from .move_device import LiveUpdatesReadbackProgressbar
from .utils import ScanRequestMixin, check_alarms

if TYPE_CHECKING:
    from bec_lib import messages
    from bec_lib.client import BECClient
    from bec_lib.queue_items import QueueItem

logger = bec_logger.logger


class IPythonLiveUpdates:
    def __init__(self, client: BECClient) -> None:
        """Class to handle live updates for IPython, also works in Jupyterlab.

        Args:
            client (BECClient): The BECClient instance.
        """
        self.client = client
        self._interrupted_request = None
        self._active_callback = None
        self._processed_instructions = 0
        self._active_request = None
        self._user_callback = None
        self._request_block_index = collections.defaultdict(lambda: 0)
        self._request_block_id = None
        self.print_table_data = True
        self._current_queue = None

    def _process_report_instructions(self, report_instructions: list) -> None:
        """Process instructions for the live updates.

        Args:
            report_instructions (list): The list of report instructions.
        """
        scan_type = self._active_request.content["scan_type"]
        if scan_type in ["open_scan_def", "close_scan_def"]:
            self._process_instruction({"table_wait": 0})
            return
        if scan_type == "close_scan_group":
            return

        if not report_instructions:
            return

        remaining_report_instructions = report_instructions[self._processed_instructions :]
        if not remaining_report_instructions:
            return

        for instr in remaining_report_instructions:
            self._process_instruction(instr)
            self._processed_instructions += 1

    def _process_instruction(self, instr: dict):
        """Process the received instruction based on scan_report_type.

        Args:
            instr (dict): The instruction to process.
        """
        scan_report_type = list(instr.keys())[0]
        scan_def_id = self.client.scans._scan_def_id
        if scan_def_id is None:
            if scan_report_type == "readback":
                LiveUpdatesReadbackProgressbar(
                    self.client,
                    report_instruction=instr,
                    request=self._active_request,
                    callbacks=self._user_callback,
                ).run()
            elif scan_report_type == "table_wait":
                LiveUpdatesTable(
                    self.client,
                    report_instruction=instr,
                    request=self._active_request,
                    callbacks=self._user_callback,
                    print_table_data=self.print_table_data,
                ).run()
            elif scan_report_type == "scan_progress":
                LiveUpdatesScanProgress(
                    self.client,
                    report_instruction=instr,
                    request=self._active_request,
                    callbacks=self._user_callback,
                ).run()
            else:
                raise ValueError(f"Unknown scan report type: {scan_report_type}")
        else:
            if self._active_callback:
                if scan_report_type == "readback":
                    LiveUpdatesReadbackProgressbar(
                        self.client,
                        report_instruction=instr,
                        request=self._active_request,
                        callbacks=self._user_callback,
                    ).run()
                else:
                    self._active_callback.resume(
                        request=self._active_request,
                        report_instruction=instr,
                        callbacks=self._user_callback,
                    )

                return

            self._active_callback = LiveUpdatesTable(
                self.client,
                report_instruction=instr,
                request=self._active_request,
                callbacks=self._user_callback,
                print_table_data=self.print_table_data,
            )
            self._active_callback.run()

    def _available_req_blocks(self, queue: QueueItem, request: messages.ScanQueueMessage):
        """Get the available request blocks.

        Args:
            queue (QueueItem): The queue item.
            request (messages.ScanQueueMessage): The request message.
        """
        available_blocks = [
            req_block
            for req_block in queue.request_blocks
            if req_block["RID"] == request.metadata["RID"]
        ]
        return available_blocks

    def process_request(
        self, request: messages.ScanQueueMessage, scan_report_type: str, callbacks: any
    ) -> None:
        """Process the request and report instructions."""
        # pylint: disable=protected-access
        try:
            with self.client._sighandler:
                # pylint: disable=protected-access
                self._active_request = request
                self._user_callback = callbacks
                scan_request = ScanRequestMixin(self.client, request.metadata["RID"])
                scan_request.wait()

                # get the corresponding queue item
                while not scan_request.request_storage.storage[-1].queue:
                    time.sleep(0.01)

                self._current_queue = queue = scan_request.request_storage.storage[-1].queue
                self._request_block_id = req_id = self._active_request.metadata.get("RID")

                while queue.status not in ["COMPLETED", "ABORTED", "HALTED"]:
                    if self._process_queue(queue, request, req_id):
                        break

                available_blocks = self._available_req_blocks(queue, request)
                req_block = available_blocks[self._request_block_index[req_id]]
                report_instructions = req_block.get("report_instructions", [])
                self._process_report_instructions(report_instructions)

            self._reset()

        except ScanInterruption as scan_interr:
            self._interrupted_request = (request, scan_report_type)
            if self._current_queue and self.client._service_config.abort_on_ctrl_c:
                self._wait_for_cleanup()
            self._reset(forced=True)
            raise scan_interr

    def _wait_for_cleanup(self):
        """Wait for the scan to be cleaned up."""
        try:
            if not self._element_in_queue():
                return
            print("Waiting for the scan to be cleaned up...")
            while self._element_in_queue():
                time.sleep(0.1)
        except KeyboardInterrupt:
            self.client.queue.request_scan_halt()

    def _element_in_queue(self) -> bool:
        queue = self.client.queue.queue_storage.current_scan_queue.get("primary", {}).get(
            "info", []
        )
        if not queue:
            return False
        return self._current_queue.queue_id in queue[0].get("queue_id")

    def _process_queue(
        self, queue: QueueItem, request: messages.ScanQueueMessage, req_id: str
    ) -> bool:
        """
        Process the queue and return True if the scan is completed.

        Args:
            queue(QueueItem): The queue item to process.
            request(messages.ScanQueueMessage): The request message.
            req_id(str): The request ID.

        Returns:
            bool: True if the scan is completed.
        """
        check_alarms(self.client)
        if not queue.request_blocks or not queue.status or queue.queue_position is None:
            return False
        if queue.status == "PENDING" and queue.queue_position > 0:
            status = self.client.queue.queue_storage.current_scan_queue.get("primary", {}).get(
                "status"
            )
            print(
                "Scan is enqueued and is waiting for execution. Current position in queue:"
                f" {queue.queue_position + 1}. Queue status: {status}.",
                end="\r",
                flush=True,
            )
        available_blocks = self._available_req_blocks(queue, request)
        req_block = available_blocks[self._request_block_index[req_id]]
        if req_block["content"]["scan_type"] == "open_scan_def":
            return True

        report_instructions = req_block["report_instructions"]
        if not report_instructions:
            return False
        self._process_report_instructions(report_instructions)

        complete_rbl = len(available_blocks) == self._request_block_index[req_id] + 1
        if self._active_callback and complete_rbl:
            return True

        if not queue.active_request_block:
            return True

        return False

    def _reset(self, forced=False):
        """Reset the active request and callback.

        Args:
            forced(bool): If True, the reset is forced.
        """
        self._interrupted_request = None

        self._current_queue = None
        self._user_callback = None
        self._processed_instructions = 0
        scan_closed = forced or (self._active_request.content["scan_type"] == "close_scan_def")
        self._active_request = None

        if self.client.scans._scan_def_id and not scan_closed:
            self._request_block_index[self._request_block_id] += 1
            return

        if scan_closed:
            self._active_callback = None

    def continue_request(self):
        """Continue the interrupted request."""
        if not self._interrupted_request:
            return
        self.process_request(*self._interrupted_request, self._user_callback)



"""
Serialization module for BEC messages
"""

from __future__ import annotations

import contextlib
import gc
import inspect
import json
from abc import abstractmethod

import msgpack as msgpack_module

from bec_lib import messages as messages_module
from bec_lib import numpy_encoder
from bec_lib.logger import bec_logger
from bec_lib.messages import BECMessage, BECStatus

logger = bec_logger.logger


def encode_bec_message_v12(msg):
    if not isinstance(msg, BECMessage):
        return msg

    msg_version = 1.2
    msg_body = msgpack.dumps(msg.__dict__)
    msg_header = json.dumps({"msg_type": msg.msg_type}).encode()
    header = f"BECMSG_{msg_version}_{len(msg_header)}_{len(msg_body)}_EOH_".encode()
    return header + msg_header + msg_body


def decode_bec_message_v12(raw_bytes):
    try:
        # kept for the record:
        # offset = MsgpackSerialization.ext_type_offset_to_data[raw_bytes[0]]
        # (was not so easy to find from msgpack doc)
        if raw_bytes.startswith(b"BECMSG"):
            version = float(raw_bytes[7:10])
            if version < 1.2:
                raise RuntimeError(f"Unsupported BECMessage version {version}")
    except Exception as exception:
        raise RuntimeError("Failed to decode BECMessage") from exception

    try:
        declaration, msg_header_body = raw_bytes.split(b"_EOH_", maxsplit=1)
        _, version, header_length, _ = declaration.split(b"_")
        header = msg_header_body[: int(header_length)]
        body = msg_header_body[int(header_length) :]
        header = json.loads(header.decode())
        msg_body = msgpack.loads(body)
        msg_class = get_message_class(header.pop("msg_type"))
        msg = msg_class(**header, **msg_body)
    except Exception as exception:
        raise RuntimeError("Failed to decode BECMessage") from exception

    # shouldn't this be checked when the msg is used? or when the message is created?
    return msg


def encode_bec_status(status):
    if not isinstance(status, BECStatus):
        return status
    return status.value.to_bytes(1, "big")  # int.to_bytes


def decode_bec_status(value):
    return BECStatus(int.from_bytes(value, "big"))


def encode_set(obj):
    if isinstance(obj, set):
        return {"__msgpack__": {"type": "set", "data": list(obj)}}
    return obj


def decode_set(obj):
    if isinstance(obj, dict) and "__msgpack__" in obj and obj["__msgpack__"]["type"] == "set":
        return set(obj["__msgpack__"]["data"])
    return obj


class SerializationRegistry:
    """Registry for serialization codecs"""

    def __init__(self):
        self._encoder = []
        self._ext_decoder = {}
        self._object_hook_decoder = []

    def register_ext_type(self, encoder, decoder):
        """Register an encoder and a decoder

        The order registrations are made counts, the encoding process is done
        in the same order until a compatible encoder is found.

        Args:
            encoder: Function encoding a data into a serializable data.
            decoder: Function decoding a serialized data into a usable data.
        """
        exttype = len(self._ext_decoder)
        if exttype in self._ext_decoder:
            raise ValueError("ExtType %d already used" % exttype)
        self._encoder.append((encoder, exttype))
        self._ext_decoder[exttype] = decoder

    def register_object_hook(self, encoder, decoder):
        """Register an encoder and a decoder that can convert a python object
        into data which can be serialized by msgpack.

        Args:
            encoder: Function encoding a data into a data serializable by msgpack
            decoder: Function decoding a python structure provided by msgpack
            into an usable data.
        """
        self._encoder.append((encoder, None))
        self._object_hook_decoder.append(decoder)

    def register_numpy(self, use_list=False):
        """
        Register BEC custom numpy encoder as a codec.
        """
        if use_list:
            self.register_object_hook(
                numpy_encoder.numpy_encode_list, numpy_encoder.numpy_decode_list
            )
        else:
            self.register_object_hook(numpy_encoder.numpy_encode, numpy_encoder.numpy_decode)

    def register_bec_message(self):
        """
        Register codec for BECMessage
        """
        # order matters
        self.register_ext_type(encode_bec_status, decode_bec_status)
        self.register_ext_type(encode_bec_message_v12, decode_bec_message_v12)

    def register_set_encoder(self):
        """
        Register codec for set
        """
        self.register_object_hook(encode_set, decode_set)


class MsgpackExt(SerializationRegistry):
    """Encapsulates msgpack dumps/loads with extensions"""

    def _default(self, obj):
        for encoder, exttype in self._encoder:
            result = encoder(obj)
            if result is obj:
                # Nothing was done, assume this encoder do not support this
                # object kind
                continue
            if exttype is not None:
                return msgpack_module.ExtType(exttype, result)
            return result
        raise TypeError("Unknown type: %r" % (obj,))

    def _ext_hooks(self, code, data):
        decoder = self._ext_decoder.get(code, None)
        if decoder is not None:
            obj = decoder(data)
            return obj
        return msgpack_module.ExtType(code, data)

    def _object_hook(self, data):
        for decoder in self._object_hook_decoder:
            try:
                result = decoder(data)
            except TypeError:
                continue
            if data is not result:
                # In case the input is not the same as the output,
                # consider it found the good decoder and it worked
                break
        else:
            return data

        return result

    def dumps(self, obj):
        """Pack object `o` and return packed bytes."""
        return msgpack_module.packb(obj, default=self._default)

    def loads(self, raw_bytes, raw=False, strict_map_key=True):
        return msgpack_module.unpackb(
            raw_bytes,
            object_hook=self._object_hook,
            ext_hook=self._ext_hooks,
            raw=raw,
            strict_map_key=strict_map_key,
        )


class JsonExt(SerializationRegistry):
    """Encapsulates JSON dumps/loads with extensions"""

    def _default(self, obj):
        for encoder, _ in self._encoder:
            result = encoder(obj)
            if result is obj:
                # Nothing was done, assume this encoder does not support this
                # object kind
                continue
            return result

    def _ext_hooks(self, data):
        for decoder in self._object_hook_decoder:
            try:
                result = decoder(data)
            except TypeError:
                continue
            if data is not result:
                # In case the input is not the same as the output,
                # consider it found the good decoder and it worked
                break
        else:
            return data
        return result

    def dumps(self, obj):
        """Serialize object `obj` and return serialized JSON string."""
        return json.dumps(obj, default=self._default)

    def loads(self, json_str):
        """Deserialize JSON string `json_str` and return the deserialized object."""
        return json.loads(json_str, object_hook=self._ext_hooks)


json_ext = JsonExt()
json_ext.register_numpy(use_list=True)
json_ext.register_bec_message()
json_ext.register_set_encoder()

msgpack = MsgpackExt()
msgpack.register_numpy()
msgpack.register_bec_message()
msgpack.register_set_encoder()


class SerializationInterface:
    """Base class for message serialization"""

    @abstractmethod
    def loads(self, msg, **kwargs) -> dict:
        """load and de-serialize a message"""

    @abstractmethod
    def dumps(self, msg, **kwargs) -> str:
        """serialize a message"""


def get_message_class(msg_type: str):
    """Given a message type, tries to find the corresponding message class in the module"""
    module = messages_module
    # convert snake_style to CamelCase
    class_name = "".join(part.title() for part in msg_type.split("_"))
    try:
        # maybe as easy as that...
        klass = getattr(module, class_name)
        # belts and braces
        if getattr(klass, "msg_type") == msg_type:
            return klass
    except AttributeError:
        # try better
        module_classes = inspect.getmembers(module, inspect.isclass)
        for class_name, klass in module_classes:
            try:
                klass_msg_type = getattr(klass, "msg_type")
            except AttributeError:
                continue
            else:
                if msg_type == klass_msg_type:
                    return klass


@contextlib.contextmanager
def pause_gc():
    """Pause the garbage collector while doing a lot of allocations, to prevent
    intempestive collect in case of big messages or if a lot of strings allocated;
    this follows the advice here: https://github.com/msgpack/msgpack-python?tab=readme-ov-file#performance-tips

    Maybe should be limited to big messages? Didn't evaluated the cost of pausing/re-enabling the GC
    """
    gc.disable()
    try:
        yield
    finally:
        gc.enable()


class MsgpackSerialization(SerializationInterface):
    """Message serialization using msgpack encoding"""

    ext_type_offset_to_data = {199: 3, 200: 4, 201: 6}

    @staticmethod
    def loads(msg) -> dict:
        with pause_gc():
            try:
                msg = msgpack.loads(msg)
            except Exception as exception:
                raise RuntimeError("Failed to decode BECMessage") from exception
            else:
                if isinstance(msg, BECMessage):
                    if msg.msg_type == "bundle_message":
                        return msg.messages
                return msg

    @staticmethod
    def dumps(msg, version=None) -> str:
        if version is None or version == 1.2:
            return msgpack.dumps(msg)
        raise RuntimeError(f"Unsupported BECMessage version {version}.")



# pylint: skip-file
import uuid
from unittest import mock

import pytest

from bec_lib import MessageEndpoints, messages
from bec_lib.tests.utils import ConnectorMock
from bec_server.scan_server.errors import DeviceMessageError, ScanAbortion
from bec_server.scan_server.scan_assembler import ScanAssembler
from bec_server.scan_server.scan_queue import (
    InstructionQueueItem,
    InstructionQueueStatus,
    QueueManager,
    RequestBlock,
    RequestBlockQueue,
    ScanQueue,
)
from bec_server.scan_server.scan_worker import ScanWorker
from bec_server.scan_server.tests.fixtures import scan_server_mock


@pytest.fixture
def scan_worker_mock(scan_server_mock) -> ScanWorker:
    scan_server_mock.device_manager.connector = mock.MagicMock()
    scan_worker = ScanWorker(parent=scan_server_mock)
    yield scan_worker


class RequestBlockQueueMock(RequestBlockQueue):
    request_blocks = []
    _scan_id = []

    @property
    def scan_id(self):
        return self._scan_id

    def append(self, msg):
        pass


class InstructionQueueMock(InstructionQueueItem):
    def __init__(self, parent: ScanQueue, assembler: ScanAssembler, worker: ScanWorker) -> None:
        super().__init__(parent, assembler, worker)
        self.queue = RequestBlockQueueMock(self, assembler)
        # self.queue.active_rb = []
        self.idx = 1

    def append_scan_request(self, msg):
        self.scan_msgs.append(msg)
        self.queue.append(msg)

    def __next__(self):
        if (
            self.status
            in [
                InstructionQueueStatus.RUNNING,
                InstructionQueueStatus.DEFERRED_PAUSE,
                InstructionQueueStatus.PENDING,
            ]
            and self.idx < 5
        ):
            self.idx += 1
            return "instr_status"

        else:
            raise StopIteration

        # while self.status == InstructionQueueStatus.PAUSED:
        #     return "instr_paused"

        # return "instr"


@pytest.mark.parametrize(
    "instruction,devices",
    [
        (
            messages.DeviceInstructionMessage(
                device="samy",
                action="wait",
                parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            ["samy"],
        ),
        (
            messages.DeviceInstructionMessage(
                device=["samx", "samy"],
                action="wait",
                parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            ["samx", "samy"],
        ),
        (
            messages.DeviceInstructionMessage(
                device="",
                action="wait",
                parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            ["samx", "samy"],
        ),
        (
            messages.DeviceInstructionMessage(
                device="",
                action="wait",
                parameter={"type": "move", "group": "primary", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            ["samx", "samy"],
        ),
        (
            messages.DeviceInstructionMessage(
                device="",
                action="wait",
                parameter={"type": "move", "group": "nogroup", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            ["samx", "samy"],
        ),
    ],
)
def test_get_devices_from_instruction(scan_worker_mock, instruction, devices):
    worker = scan_worker_mock
    worker.scan_motors = devices
    worker.readout_priority.update({"monitored": devices})

    returned_devices = worker._get_devices_from_instruction(instruction)

    if not instruction.content.get("device"):
        group = instruction.content["parameter"].get("group")
        if group == "primary":
            assert (
                set(
                    dev.name
                    for dev in worker.device_manager.devices.monitored_devices(worker.scan_motors)
                ).difference(set(dev.name for dev in returned_devices))
                == set()
            )
        elif group == "scan_motor":
            assert returned_devices == devices
        else:
            assert returned_devices == []
    else:
        assert returned_devices == [worker.device_manager.devices[dev] for dev in devices]


@pytest.mark.parametrize(
    "instructions",
    [
        (
            messages.DeviceInstructionMessage(
                device="samx",
                action="set",
                parameter={"value": 10, "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            )
        ),
        messages.DeviceInstructionMessage(
            device="samx",
            action="set",
            parameter={"value": 10, "wait_group": "scan_motor"},
            metadata={"readout_priority": "monitored", "DIID": None},
        ),
    ],
)
def test_add_wait_group(scan_worker_mock, instructions):
    worker = scan_worker_mock
    if instructions.metadata["DIID"]:
        worker._add_wait_group(instructions)
        assert worker._groups == {"scan_motor": {"samx": 3}}

        worker._groups["scan_motor"] = {"samy": 2}
        worker._add_wait_group(instructions)
        assert worker._groups == {"scan_motor": {"samy": 2, "samx": 3}}

    else:
        with pytest.raises(DeviceMessageError) as exc_info:
            worker._add_wait_group(instructions)
        assert exc_info.value.args[0] == "Device message metadata does not contain a DIID entry."


def test_add_wait_group_to_existing_wait_group(scan_worker_mock):
    instr1 = messages.DeviceInstructionMessage(
        device="samx",
        action="set",
        parameter={"value": 10, "wait_group": "scan_motor"},
        metadata={"readout_priority": "monitored", "DIID": 3},
    )
    instr2 = messages.DeviceInstructionMessage(
        device="samx",
        action="set",
        parameter={"value": 10, "wait_group": "scan_motor"},
        metadata={"readout_priority": "monitored", "DIID": 4},
    )
    worker = scan_worker_mock
    worker._add_wait_group(instr1)
    worker._add_wait_group(instr2)
    assert worker._groups == {"scan_motor": {"samx": 4}}


@pytest.mark.parametrize(
    "instructions,wait_type",
    [
        (
            messages.DeviceInstructionMessage(
                device="samy",
                action="wait",
                parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            "move",
        ),
        (
            messages.DeviceInstructionMessage(
                device="samy",
                action="wait",
                parameter={"type": "read", "group": "scan_motor", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            "read",
        ),
        (
            messages.DeviceInstructionMessage(
                device="samy",
                action="wait",
                parameter={"type": "trigger", "group": "scan_motor", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            "trigger",
        ),
        (
            messages.DeviceInstructionMessage(
                device="samy",
                action="wait",
                parameter={"type": None, "group": "scan_motor", "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 3},
            ),
            None,
        ),
    ],
)
def test_wait_for_devices(scan_worker_mock, instructions, wait_type):
    worker = scan_worker_mock

    with mock.patch.object(worker, "_wait_for_idle") as idle_mock:
        with mock.patch.object(worker, "_wait_for_read") as read_mock:
            with mock.patch.object(worker, "_wait_for_trigger") as trigger_mock:
                if wait_type:
                    worker.wait_for_devices(instructions)

                if wait_type == "move":
                    idle_mock.assert_called_once_with(instructions)
                elif wait_type == "read":
                    read_mock.assert_called_once_with(instructions)
                elif wait_type == "trigger":
                    trigger_mock.assert_called_once_with(instructions)
                else:
                    with pytest.raises(DeviceMessageError) as exc_info:
                        worker.wait_for_devices(instructions)
                    assert exc_info.value.args[0] == "Unknown wait command"


@pytest.mark.parametrize(
    "instructions",
    [
        (
            messages.DeviceInstructionMessage(
                device="samx",
                action="complete",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 3},
            )
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="complete",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 3},
            )
        ),
        (
            messages.DeviceInstructionMessage(
                device=["samx", "samy"],
                action="complete",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 3},
            )
        ),
    ],
)
def test_complete_devices(scan_worker_mock, instructions):
    worker = scan_worker_mock
    with mock.patch.object(worker, "_wait_for_status") as wait_for_status_mock:
        with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
            worker.complete_devices(instructions)
            if instructions.content["device"]:
                devices = instructions.content["device"]
                if isinstance(devices, str):
                    devices = [devices]
            else:
                devices = [dev.name for dev in worker.device_manager.devices.enabled_devices]

            wait_for_status_mock.assert_called_once_with(devices, instructions.metadata)
            send_mock.assert_called_once_with(
                MessageEndpoints.device_instructions(),
                messages.DeviceInstructionMessage(
                    device=devices, action="complete", parameter={}, metadata=instructions.metadata
                ),
            )


@pytest.mark.parametrize(
    "instructions",
    [
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="pre_scan",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 3},
            )
        )
    ],
)
def test_pre_scan(scan_worker_mock, instructions):
    worker = scan_worker_mock
    with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
        with mock.patch.object(worker, "_wait_for_status") as wait_for_status_mock:
            worker.pre_scan(instructions)
            devices = [dev.name for dev in worker.device_manager.devices.enabled_devices]

            wait_for_status_mock.assert_called_once_with(devices, instructions.metadata)
            send_mock.assert_called_once_with(
                MessageEndpoints.device_instructions(),
                messages.DeviceInstructionMessage(
                    device=devices, action="pre_scan", parameter={}, metadata=instructions.metadata
                ),
            )


@pytest.mark.parametrize(
    "device_status,devices,instr,abort",
    [
        (
            [
                messages.DeviceReqStatusMessage(
                    device="samx",
                    success=True,
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 3,
                        "scan_id": "scan_id",
                        "RID": "requestID",
                    },
                )
            ],
            [("samx", 4)],
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            False,
        ),
        (
            [
                messages.DeviceReqStatusMessage(
                    device="samx",
                    success=False,
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 3,
                        "scan_id": "scan_id",
                        "RID": "request",
                    },
                )
            ],
            [("samx", 4)],
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            False,
        ),
        (
            [
                messages.DeviceReqStatusMessage(
                    device="samx",
                    success=False,
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 4,
                        "scan_id": "scan_id",
                        "RID": "requestID",
                    },
                )
            ],
            [("samx", 4)],
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            True,
        ),
        (
            [
                messages.DeviceReqStatusMessage(
                    device="samx",
                    success=False,
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 3,
                        "scan_id": "scan_id",
                        "RID": "requestID",
                    },
                )
            ],
            [("samx", 4)],
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            False,
        ),
    ],
)
def test_check_for_failed_movements(scan_worker_mock, device_status, devices, instr, abort):
    worker = scan_worker_mock
    worker.device_manager.connector = ConnectorMock()
    if abort:
        with pytest.raises(ScanAbortion):
            worker.device_manager.connector._get_buffer[
                MessageEndpoints.device_readback("samx").endpoint
            ] = messages.DeviceMessage(signals={"samx": {"value": 4}}, metadata={})
            worker._check_for_failed_movements(device_status, devices, instr)
    else:
        worker._check_for_failed_movements(device_status, devices, instr)


@pytest.mark.parametrize(
    "msg1,msg2,req_msg",
    [
        (
            messages.DeviceInstructionMessage(
                device="samx",
                action="set",
                parameter={"value": 10, "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceReqStatusMessage(
                device="samx",
                success=False,
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
        ),
        (
            messages.DeviceInstructionMessage(
                device="samx",
                action="set",
                parameter={"value": 10, "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceReqStatusMessage(
                device="samx",
                success=True,
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
        ),
        (
            messages.DeviceInstructionMessage(
                device="samx",
                action="set",
                parameter={"value": 10, "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceReqStatusMessage(
                device="samx",
                success=True,
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 5,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
        ),
    ],
)
def test_wait_for_idle(scan_worker_mock, msg1, msg2, req_msg: messages.DeviceReqStatusMessage):
    worker = scan_worker_mock
    worker.device_manager.connector = ConnectorMock()

    with mock.patch.object(
        worker.validate, "get_device_status", return_value=[req_msg]
    ) as device_status:
        worker.device_manager.connector._get_buffer[
            MessageEndpoints.device_readback("samx").endpoint
        ] = messages.DeviceMessage(signals={"samx": {"value": 4}}, metadata={})

        worker._add_wait_group(msg1)
        if req_msg.content["success"]:
            worker._wait_for_idle(msg2)
        else:
            with pytest.raises(ScanAbortion):
                worker._wait_for_idle(msg2)


@pytest.mark.parametrize(
    "msg1,msg2,req_msg",
    [
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="set",
                parameter={"value": 10, "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            messages.DeviceStatusMessage(
                device="samx",
                status=0,
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
        )
    ],
)
def test_wait_for_read(scan_worker_mock, msg1, msg2, req_msg: messages.DeviceReqStatusMessage):
    worker = scan_worker_mock
    worker.device_manager.connector = ConnectorMock()

    with mock.patch.object(
        worker.validate, "get_device_status", return_value=[req_msg]
    ) as device_status:
        with mock.patch.object(worker, "_check_for_interruption") as interruption_mock:
            assert worker._groups == {}
            worker._groups["scan_motor"] = {"samx": 3, "samy": 4}
            worker.device_manager.connector._get_buffer[
                MessageEndpoints.device_readback("samx").endpoint
            ] = messages.DeviceMessage(signals={"samx": {"value": 4}}, metadata={})
            worker._add_wait_group(msg1)
            worker._wait_for_read(msg2)
            assert worker._groups == {"scan_motor": {"samy": 4}}
            interruption_mock.assert_called_once()


@pytest.mark.parametrize(
    "instr",
    [
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="set",
                parameter={"time": 0.1},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            )
        )
    ],
)
def test_wait_for_trigger(scan_worker_mock, instr):
    worker = scan_worker_mock
    worker._last_trigger = instr

    with mock.patch.object(worker.validate, "get_device_status") as status_mock:
        with mock.patch.object(worker, "_check_for_interruption") as interruption_mock:
            status_mock.return_value = [
                messages.DeviceReqStatusMessage(
                    device="eiger",
                    success=True,
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 3,
                        "scan_id": "scan_id",
                        "RID": "requestID",
                    },
                )
            ]
            worker._wait_for_trigger(instr)
            status_mock.assert_called_once_with(MessageEndpoints.device_req_status, ["eiger"])
            interruption_mock.assert_called_once()


def test_wait_for_stage(scan_worker_mock):
    worker = scan_worker_mock
    devices = ["samx", "samy"]
    with mock.patch.object(worker.validate, "get_device_status") as status_mock:
        with mock.patch.object(worker, "_check_for_interruption") as interruption_mock:
            worker._wait_for_stage(True, devices, {})
            status_mock.assert_called_once_with(MessageEndpoints.device_staged, devices)
            interruption_mock.assert_called_once()


def test_wait_for_device_server(scan_worker_mock):
    worker = scan_worker_mock
    with mock.patch.object(worker.parent, "wait_for_service") as service_mock:
        worker._wait_for_device_server()
        service_mock.assert_called_once_with("DeviceServer")


@pytest.mark.parametrize(
    "instr",
    [
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="set",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            )
        )
    ],
)
def test_set_devices(scan_worker_mock, instr):
    worker = scan_worker_mock
    with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
        worker.set_devices(instr)
        send_mock.assert_called_once_with(MessageEndpoints.device_instructions(), instr)


@pytest.mark.parametrize(
    "instr",
    [
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="trigger",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            )
        )
    ],
)
def test_trigger_devices(scan_worker_mock, instr):
    worker = scan_worker_mock
    with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
        worker.trigger_devices(instr)
        devices = [
            dev.name for dev in worker.device_manager.devices.get_software_triggered_devices()
        ]

        send_mock.assert_called_once_with(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=devices,
                action="trigger",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
        )


@pytest.mark.parametrize(
    "instr",
    [
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="trigger",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            )
        )
    ],
)
def test_send_rpc(scan_worker_mock, instr):
    worker = scan_worker_mock
    with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
        worker.send_rpc(instr)
        send_mock.assert_called_once_with(MessageEndpoints.device_instructions(), instr)


@pytest.mark.parametrize(
    "instr",
    [
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="read",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            )
        ),
        (
            messages.DeviceInstructionMessage(
                device=[],
                action="read",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            )
        ),
    ],
)
def test_read_devices(scan_worker_mock, instr):
    worker = scan_worker_mock
    instr_devices = instr.content["device"]
    if instr_devices is None:
        instr_devices = []
    worker.readout_priority.update({"monitored": instr_devices})
    devices = [dev.name for dev in worker._get_devices_from_instruction(instr)]
    with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
        worker.read_devices(instr)

        if instr.content.get("device"):
            send_mock.assert_called_once_with(
                MessageEndpoints.device_instructions(),
                messages.DeviceInstructionMessage(
                    device=["samx"],
                    action="read",
                    parameter=instr.content["parameter"],
                    metadata=instr.metadata,
                ),
            )
        else:
            send_mock.assert_called_once_with(
                MessageEndpoints.device_instructions(),
                messages.DeviceInstructionMessage(
                    device=devices,
                    action="read",
                    parameter=instr.content["parameter"],
                    metadata=instr.metadata,
                ),
            )


@pytest.mark.parametrize(
    "instr, devices, parameter, metadata",
    [
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="trigger",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            ["samx"],
            {"value": 10, "wait_group": "scan_motor", "time": 30},
            {"readout_priority": "monitored", "DIID": 3, "scan_id": "scan_id", "RID": "requestID"},
        )
    ],
)
def test_kickoff_devices(scan_worker_mock, instr, devices, parameter, metadata):
    worker = scan_worker_mock
    with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
        worker.kickoff_devices(instr)
        send_mock.assert_called_once_with(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=devices, action="kickoff", parameter=parameter, metadata=metadata
            ),
        )


@pytest.mark.parametrize(
    "instr, devices",
    [
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="trigger",
                parameter={"value": 10, "wait_group": "scan_motor", "time": 30},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 3,
                    "scan_id": "scan_id",
                    "RID": "requestID",
                },
            ),
            None,
        )
    ],
)
def test_publish_readback(scan_worker_mock, instr, devices):
    worker = scan_worker_mock
    with mock.patch.object(worker, "_get_readback", return_value=[{}]) as get_readback:
        with mock.patch.object(worker.device_manager, "connector") as connector_mock:
            worker._publish_readback(instr)

            get_readback.assert_called_once_with(["samx"])
            pipe = connector_mock.pipeline()
            msg = messages.DeviceMessage(signals={}, metadata=instr.metadata)
            connector_mock.set_and_publish.assert_called_once_with(
                MessageEndpoints.device_read("samx"), msg, pipe
            )


def test_get_readback(scan_worker_mock):
    worker = scan_worker_mock
    devices = ["samx"]
    with mock.patch.object(worker.device_manager, "connector") as connector_mock:
        worker._get_readback(devices)
        pipe = connector_mock.pipeline()
        connector_mock.get.assert_called_once_with(
            MessageEndpoints.device_readback("samx"), pipe=pipe
        )
        connector_mock.execute_pipeline.assert_called_once()


def test_publish_data_as_read(scan_worker_mock):
    worker = scan_worker_mock
    instr = messages.DeviceInstructionMessage(
        device=["samx"],
        action="publish_data_as_read",
        parameter={"data": {}},
        metadata={
            "readout_priority": "monitored",
            "DIID": 3,
            "scan_id": "scan_id",
            "RID": "requestID",
        },
    )
    with mock.patch.object(worker.device_manager, "connector") as connector_mock:
        worker.publish_data_as_read(instr)
        msg = messages.DeviceMessage(
            signals=instr.content["parameter"]["data"], metadata=instr.metadata
        )
        connector_mock.set_and_publish.assert_called_once_with(
            MessageEndpoints.device_read("samx"), msg
        )


def test_publish_data_as_read_multiple(scan_worker_mock):
    worker = scan_worker_mock
    data = [{"samx": {}}, {"samy": {}}]
    devices = ["samx", "samy"]
    instr = messages.DeviceInstructionMessage(
        device=devices,
        action="publish_data_as_read",
        parameter={"data": data},
        metadata={
            "readout_priority": "monitored",
            "DIID": 3,
            "scan_id": "scan_id",
            "RID": "requestID",
        },
    )
    with mock.patch.object(worker.device_manager, "connector") as connector_mock:
        worker.publish_data_as_read(instr)
        mock_calls = []
        for device, dev_data in zip(devices, data):
            msg = messages.DeviceMessage(signals=dev_data, metadata=instr.metadata)
            mock_calls.append(mock.call(MessageEndpoints.device_read(device), msg))
        assert connector_mock.set_and_publish.mock_calls == mock_calls


def test_check_for_interruption(scan_worker_mock):
    worker = scan_worker_mock
    worker.status = InstructionQueueStatus.STOPPED
    with pytest.raises(ScanAbortion) as exc_info:
        worker._check_for_interruption()


@pytest.mark.parametrize(
    "instr, corr_num_points, scan_id",
    [
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="open_scan",
                parameter={"num_points": 150, "scan_motors": ["samx", "samy"]},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 18,
                    "scan_id": "12345",
                    "scan_def_id": 100,
                    "point_id": 50,
                    "RID": 11,
                },
            ),
            201,
            False,
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="open_scan",
                parameter={"num_points": 150},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 18,
                    "scan_id": "12345",
                    "RID": 11,
                },
            ),
            150,
            True,
        ),
    ],
)
def test_open_scan(scan_worker_mock, instr, corr_num_points, scan_id):
    worker = scan_worker_mock

    if not scan_id:
        assert worker.scan_id == None
    else:
        worker.scan_id = 111
        worker.scan_motors = ["bpm4i"]

    if "point_id" in instr.metadata:
        worker.max_point_id = instr.metadata["point_id"]

    assert worker.parent.connector.get(MessageEndpoints.scan_number()) == None

    with mock.patch.object(worker, "current_instruction_queue_item") as queue_mock:
        with mock.patch.object(worker, "_initialize_scan_info") as init_mock:
            with mock.patch.object(worker.scan_report_instructions, "append") as instr_append_mock:
                with mock.patch.object(worker, "_send_scan_status") as send_mock:
                    with mock.patch.object(
                        worker.current_instruction_queue_item.parent.queue_manager,
                        "send_queue_status",
                    ) as queue_status_mock:
                        active_rb = queue_mock.active_request_block
                        active_rb.scan_report_instructions = []
                        worker.open_scan(instr)

                        if not scan_id:
                            assert worker.scan_id == instr.metadata.get("scan_id")
                            assert worker.scan_motors == [
                                worker.device_manager.devices["samx"],
                                worker.device_manager.devices["samy"],
                            ]
                        else:
                            assert worker.scan_id == 111
                            assert worker.scan_motors == ["bpm4i"]
                        init_mock.assert_called_once_with(active_rb, instr, corr_num_points)
                        assert active_rb.scan_report_instructions == [
                            {"table_wait": corr_num_points}
                        ]
                        queue_status_mock.assert_called_once()
                        send_mock.assert_called_once_with("open")


@pytest.mark.parametrize(
    "msg",
    [
        messages.ScanQueueMessage(
            scan_type="grid_scan",
            parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}, "num_points": 100},
            queue="primary",
            metadata={"RID": "something"},
        )
    ],
)
def test_initialize_scan_info(scan_worker_mock, msg):
    worker = scan_worker_mock
    scan_server = scan_worker_mock.parent
    rb = RequestBlock(msg, assembler=ScanAssembler(parent=scan_server))
    assert rb.metadata == {"RID": "something"}

    with mock.patch.object(worker, "current_instruction_queue_item"):
        worker.scan_motors = ["samx"]
        worker.readout_priority = {
            "monitored": ["samx"],
            "baseline": [],
            "async": [],
            "continuous": [],
            "on_request": [],
        }
        worker._initialize_scan_info(rb, msg, msg.content["parameter"].get("num_points"))

        assert worker.current_scan_info["RID"] == "something"
        assert worker.current_scan_info["scan_number"] == 2
        assert worker.current_scan_info["dataset_number"] == 3
        assert worker.current_scan_info["scan_report_hint"] == rb.scan.scan_report_hint
        assert worker.current_scan_info["scan_report_devices"] == rb.scan.scan_report_devices
        assert worker.current_scan_info["num_points"] == 100
        assert worker.current_scan_info["scan_msgs"] == []
        assert worker.current_scan_info["monitor_sync"] == "bec"
        assert worker.current_scan_info["frames_per_trigger"] == 1
        assert worker.current_scan_info["args"] == {"samx": (-5, 5, 3)}
        assert worker.current_scan_info["kwargs"] == {}
        assert "samx" in worker.current_scan_info["readout_priority"]["monitored"]
        assert "samy" in worker.current_scan_info["readout_priority"]["baseline"]


@pytest.mark.parametrize(
    "msg,scan_id,max_point_id,exp_num_points",
    [
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="close_scan",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 18, "scan_id": "12345"},
            ),
            "12345",
            19,
            20,
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="close_scan",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 18, "scan_id": "12345"},
            ),
            "0987",
            200,
            19,
        ),
    ],
)
def test_close_scan(scan_worker_mock, msg, scan_id, max_point_id, exp_num_points):
    worker = scan_worker_mock
    worker.scan_id = scan_id
    worker.current_scan_info["num_points"] = 19

    reset = bool(worker.scan_id == msg.metadata["scan_id"])
    with mock.patch.object(worker, "_send_scan_status") as send_scan_status_mock:
        worker.close_scan(msg, max_point_id=max_point_id)
        if reset:
            send_scan_status_mock.assert_called_with("closed")
            assert worker.scan_id == None
        else:
            assert worker.scan_id == scan_id
    assert worker.current_scan_info["num_points"] == exp_num_points


@pytest.mark.parametrize(
    "msg",
    [
        messages.DeviceInstructionMessage(
            device=None,
            action="stage",
            parameter={},
            metadata={"readout_priority": "async", "DIID": 18, "scan_id": "12345"},
        )
    ],
)
def test_stage_device(scan_worker_mock, msg):
    worker = scan_worker_mock
    worker.device_manager.devices["eiger"]._config["readoutPriority"] = "async"
    worker.device_manager.devices["flyer_sim"]._config["readoutPriority"] = "on_request"

    with mock.patch.object(worker, "_wait_for_stage") as wait_mock:
        with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
            worker.stage_devices(msg)
            on_request_device_names = [
                dev.name for dev in worker.device_manager.devices.on_request_devices()
            ]

            async_devices = worker.device_manager.devices.async_devices()
            async_device_names = [dev.name for dev in async_devices]
            excluded_devices = async_devices
            excluded_devices.extend(worker.device_manager.devices.on_request_devices())
            excluded_devices.extend(worker.device_manager.devices.continuous_devices())
            devices = [
                dev.name
                for dev in worker.device_manager.devices.enabled_devices
                if dev not in excluded_devices
            ]

            for dev in [
                *worker.device_manager.devices.monitored_devices(),
                *worker.device_manager.devices.baseline_devices(),
                *worker.device_manager.devices.async_devices(),
            ]:
                assert dev.name in worker._staged_devices
            for async_dev in async_devices:
                assert (
                    mock.call(
                        MessageEndpoints.device_instructions(),
                        messages.DeviceInstructionMessage(
                            device=async_dev.name,
                            action="stage",
                            parameter=msg.content["parameter"],
                            metadata=msg.metadata,
                        ),
                    )
                    in send_mock.mock_calls
                )
            assert (
                mock.call(
                    MessageEndpoints.device_instructions(),
                    messages.DeviceInstructionMessage(
                        device=devices,
                        action="stage",
                        parameter=msg.content["parameter"],
                        metadata=msg.metadata,
                    ),
                )
                in send_mock.mock_calls
            )
            assert (
                mock.call(staged=True, devices=devices, metadata=msg.metadata)
                in wait_mock.mock_calls
            )
            assert (
                mock.call(staged=True, devices=async_device_names, metadata=msg.metadata)
                in wait_mock.mock_calls
            )
            for dev in on_request_device_names:
                assert dev not in worker._staged_devices


@pytest.mark.parametrize(
    "msg, devices, parameter, metadata, cleanup",
    [
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="close_scan",
                parameter={"parameter": "param"},
                metadata={"readout_priority": "monitored", "DIID": 18, "scan_id": "12345"},
            ),
            ["samx"],
            {"parameter": "param"},
            {"readout_priority": "monitored", "DIID": 18, "scan_id": "12345"},
            False,
        ),
        (None, None, {}, {}, False),
        (None, None, {}, {}, True),
    ],
)
def test_unstage_device(scan_worker_mock, msg, devices, parameter, metadata, cleanup):
    worker = scan_worker_mock
    if not devices:
        devices = [dev.name for dev in worker.device_manager.devices.enabled_devices]

    with mock.patch.object(worker.device_manager.connector, "send") as send_mock:
        with mock.patch.object(worker, "_wait_for_stage") as wait_mock:
            worker.unstage_devices(msg, devices, cleanup)

            send_mock.assert_called_once_with(
                MessageEndpoints.device_instructions(),
                messages.DeviceInstructionMessage(
                    device=devices, action="unstage", parameter=parameter, metadata=metadata
                ),
            )
            if cleanup:
                wait_mock.assert_not_called()
            else:
                wait_mock.assert_called_once_with(staged=False, devices=devices, metadata=metadata)


@pytest.mark.parametrize("status,expire", [("open", None), ("closed", 1800), ("aborted", 1800)])
def test_send_scan_status(scan_worker_mock, status, expire):
    worker = scan_worker_mock
    worker.device_manager.connector = ConnectorMock()
    worker.current_scan_id = str(uuid.uuid4())
    worker._send_scan_status(status)
    scan_info_msgs = [
        msg
        for msg in worker.device_manager.connector.message_sent
        if msg["queue"]
        == MessageEndpoints.public_scan_info(scan_id=worker.current_scan_id).endpoint
    ]
    assert len(scan_info_msgs) == 1
    assert scan_info_msgs[0]["expire"] == expire


@pytest.mark.parametrize("abortion", [False, True])
def test_process_instructions(scan_worker_mock, abortion):
    worker = scan_worker_mock
    scan_server = scan_worker_mock.parent
    scan_queue = ScanQueue(QueueManager(scan_server))
    queue = InstructionQueueMock(
        parent=scan_queue, assembler=ScanAssembler(parent=scan_server), worker=worker
    )

    with mock.patch.object(worker, "_wait_for_device_server") as wait_mock:
        with mock.patch.object(worker, "reset") as reset_mock:
            with mock.patch.object(worker, "_check_for_interruption") as interruption_mock:
                with mock.patch.object(queue.queue, "active_rb") as rb_mock:
                    with mock.patch.object(worker, "_instruction_step") as step_mock:
                        if abortion:
                            interruption_mock.side_effect = ScanAbortion
                            with pytest.raises(ScanAbortion) as exc_info:
                                worker._process_instructions(queue)
                        else:
                            worker._process_instructions(queue)

                        assert worker.max_point_id == 0
                        wait_mock.assert_called_once()

                        if not abortion:
                            assert interruption_mock.call_count == 4
                            assert worker._exposure_time == getattr(rb_mock.scan, "exp_time", None)
                            assert step_mock.call_count == 4
                            assert queue.is_active == False
                            assert queue.status == InstructionQueueStatus.COMPLETED
                            assert worker.current_instruction_queue_item == None
                            reset_mock.assert_called_once()

                        else:
                            assert worker._groups == {}
                            assert queue.stopped == True
                            assert interruption_mock.call_count == 1
                            assert queue.is_active == True
                            assert queue.status == InstructionQueueStatus.PENDING
                            assert worker.current_instruction_queue_item == queue


@pytest.mark.parametrize(
    "msg,method",
    [
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="open_scan",
                parameter={"readout_priority": {"monitored": [], "baseline": [], "on_request": []}},
                metadata={"readout_priority": "monitored", "DIID": 18, "scan_id": "12345"},
            ),
            "open_scan",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="close_scan",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 18, "scan_id": "12345"},
            ),
            "close_scan",
        ),
        (
            messages.DeviceInstructionMessage(
                device=["samx"],
                action="wait",
                parameter={"type": "move", "wait_group": "scan_motor"},
                metadata={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "scan_id": "12345",
                    "RID": "123456",
                },
            ),
            "wait_for_devices",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="trigger",
                parameter={"group": "trigger"},
                metadata={"readout_priority": "monitored", "DIID": 20, "point_id": 0},
            ),
            "trigger_devices",
        ),
        (
            messages.DeviceInstructionMessage(
                device="samx",
                action="set",
                parameter={"value": 1.3681828686580249, "wait_group": "scan_motor"},
                metadata={"readout_priority": "monitored", "DIID": 24},
            ),
            "set_devices",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="read",
                parameter={"group": "primary", "wait_group": "readout_primary"},
                metadata={"readout_priority": "monitored", "DIID": 30, "point_id": 1},
            ),
            "read_devices",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="stage",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 17},
            ),
            "stage_devices",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="unstage",
                parameter={},
                metadata={"readout_priority": "monitored", "DIID": 17},
            ),
            "unstage_devices",
        ),
        (
            messages.DeviceInstructionMessage(
                device="samx",
                action="rpc",
                parameter={
                    "device": "lsamy",
                    "func": "readback.get",
                    "rpc_id": "61a7376c-36cf-41af-94b1-76c1ba821d47",
                    "args": [],
                    "kwargs": {},
                },
                metadata={"readout_priority": "monitored", "DIID": 9},
            ),
            "send_rpc",
        ),
        (
            messages.DeviceInstructionMessage(
                device="samx", action="kickoff", parameter={}, metadata={}
            ),
            "kickoff_devices",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None,
                action="baseline_reading",
                parameter={},
                metadata={"readout_priority": "baseline", "DIID": 15},
            ),
            "baseline_reading",
        ),
        (
            messages.DeviceInstructionMessage(device=None, action="close_scan_def", parameter={}),
            "close_scan",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None, action="publish_data_as_read", parameter={}
            ),
            "publish_data_as_read",
        ),
        (
            messages.DeviceInstructionMessage(
                device=None, action="scan_report_instruction", parameter={}
            ),
            "process_scan_report_instruction",
        ),
        (
            messages.DeviceInstructionMessage(device=None, action="pre_scan", parameter={}),
            "pre_scan",
        ),
        (
            messages.DeviceInstructionMessage(device=None, action="complete", parameter={}),
            "complete_devices",
        ),
    ],
)
def test_instruction_step(scan_worker_mock, msg, method):
    worker = scan_worker_mock
    with mock.patch(
        f"bec_server.scan_server.scan_worker.ScanWorker.{method}"
    ) as instruction_method:
        worker._instruction_step(msg)
        instruction_method.assert_called_once()


def test_reset(scan_worker_mock):
    worker = scan_worker_mock
    worker._gropus = 1
    worker.current_scan_id = 1
    worker.current_scan_info = 1
    worker.scan_id = 1
    worker.interception_msg = 1
    worker.scan_motors = 1

    worker.reset()

    assert worker._groups == {}
    assert worker.current_scan_id == ""
    assert worker.current_scan_info == {}
    assert worker.scan_id == None
    assert worker.interception_msg == None
    assert worker.scan_motors == []


def test_cleanup(scan_worker_mock):
    worker = scan_worker_mock
    with mock.patch.object(worker, "unstage_devices") as unstage_mock:
        worker.cleanup()
        unstage_mock.assert_called_once_with(devices=list(worker._staged_devices), cleanup=True)


def test_shutdown(scan_worker_mock):
    worker = scan_worker_mock
    with mock.patch.object(worker.signal_event, "set") as set_mock:
        worker._started = mock.MagicMock()
        worker._started.is_set.return_value = True
        with mock.patch.object(worker, "join") as join_mock:
            worker.shutdown()
            set_mock.assert_called_once()
            join_mock.assert_called_once()



from unittest import mock

from bec_server.data_processing.cli.launch import main


def test_main():
    with mock.patch("bec_server.data_processing.cli.launch.argparse.ArgumentParser") as mock_parser:
        with mock.patch("bec_server.data_processing.cli.launch.ServiceConfig") as mock_config:
            with mock.patch(
                "bec_server.data_processing.dap_server.DAPServer"
            ) as mock_data_processing:
                with mock.patch(
                    "bec_server.data_processing.cli.launch.threading.Event"
                ) as mock_event:
                    main()
                    mock_parser.assert_called_once()
                    mock_config.assert_called_once()
                    mock_data_processing.assert_called_once()
                    mock_event.assert_called_once()


def test_main_shutdown():
    with mock.patch("bec_server.data_processing.cli.launch.argparse.ArgumentParser") as mock_parser:
        with mock.patch("bec_server.data_processing.cli.launch.ServiceConfig") as mock_config:
            with mock.patch(
                "bec_server.data_processing.dap_server.DAPServer"
            ) as mock_data_processing:
                with mock.patch(
                    "bec_server.data_processing.cli.launch.threading.Event"
                ) as mock_event:
                    mock_event.return_value.wait.side_effect = KeyboardInterrupt()
                    try:
                        main()
                    except KeyboardInterrupt:
                        pass
                    mock_parser.assert_called_once()
                    mock_config.assert_called_once()
                    mock_data_processing.assert_called_once()
                    mock_event.assert_called_once()
                    mock_data_processing.return_value.shutdown.assert_called_once()



from bec_lib.alarm_handler import Alarms
from bec_lib.bec_service import BECService
from bec_lib.channel_monitor import channel_monitor_launch, log_monitor_launch
from bec_lib.client import BECClient
from bec_lib.config_helper import ConfigHelper
from bec_lib.device import DeviceBase, DeviceStatus, Status
from bec_lib.devicemanager import DeviceConfigError, DeviceContainer, DeviceManagerBase
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.messages import BECStatus
from bec_lib.redis_connector import RedisConnector
from bec_lib.service_config import ServiceConfig
from bec_lib.utils import threadlocked



import os
import shutil
from unittest import mock

import pytest
import yaml

import bec_lib
from bec_lib import messages
from bec_lib.bec_errors import DeviceConfigError, ServiceConfigError
from bec_lib.config_helper import ConfigHelper

dir_path = os.path.dirname(bec_lib.__file__)


def test_load_demo_config():
    connector = mock.MagicMock()
    config_helper = ConfigHelper(connector)
    with mock.patch.object(config_helper, "update_session_with_file") as mock_update:
        config_helper.load_demo_config()
        dirpath = os.path.dirname(bec_lib.__file__)
        fpath = os.path.join(dirpath, "configs/demo_config.yaml")
        mock_update.assert_called_once_with(fpath)


def test_config_helper_update_session_with_file():
    connector = mock.MagicMock()
    config_helper = ConfigHelper(connector)
    with mock.patch.object(config_helper, "send_config_request") as mock_send_config_request:
        with mock.patch.object(
            config_helper, "_load_config_from_file"
        ) as mock_load_config_from_file:
            mock_load_config_from_file.return_value = {"test": "test"}
            config_helper._base_path_recovery = "."
            config_helper.update_session_with_file("test.yaml")
            mock_send_config_request.assert_called_once_with(action="set", config={"test": "test"})


@pytest.mark.parametrize("config_file", ["test.yaml", "test.yml"])
def test_config_helper_load_config_from_file(tmp_path, test_config_yaml_file_path, config_file):
    orig_cfg_file = test_config_yaml_file_path
    test_cfg_file = tmp_path / config_file
    shutil.copyfile(orig_cfg_file, test_cfg_file)
    connector = mock.MagicMock()
    config_helper = ConfigHelper(connector)
    config = config_helper._load_config_from_file(test_cfg_file)


def test_config_helper_save_current_session():
    connector = mock.MagicMock()

    config_helper = ConfigHelper(connector)
    connector.get.return_value = messages.AvailableResourceMessage(
        resource=[
            {
                "id": "648c817f67d3c7cd6a354e8e",
                "createdAt": "2023-06-16T15:36:31.215Z",
                "createdBy": "unknown user",
                "name": "pinz",
                "sessionId": "648c817d67d3c7cd6a354df2",
                "enabled": True,
                "readOnly": False,
                "deviceClass": "SimPositioner",
                "deviceTags": ["user motors"],
                "deviceConfig": {
                    "delay": 1,
                    "labels": "pinz",
                    "limits": [-50, 50],
                    "name": "pinz",
                    "speed": 100,
                    "tolerance": 0.01,
                    "update_frequency": 400,
                },
                "readoutPriority": "baseline",
                "onFailure": "retry",
            },
            {
                "id": "648c817f67d3c7cd6a354ec5",
                "createdAt": "2023-06-16T15:36:31.764Z",
                "createdBy": "unknown user",
                "name": "transd",
                "sessionId": "648c817d67d3c7cd6a354df2",
                "enabled": True,
                "readOnly": False,
                "deviceClass": "SimMonitor",
                "deviceTags": ["beamline"],
                "deviceConfig": {"labels": "transd", "name": "transd", "tolerance": 0.5},
                "readoutPriority": "monitored",
                "onFailure": "retry",
            },
        ]
    )
    with mock.patch("builtins.open", mock.mock_open()) as mock_open:
        config_helper.save_current_session("test.yaml")
        out_data = {
            "pinz": {
                "deviceClass": "SimPositioner",
                "deviceTags": ["user motors"],
                "enabled": True,
                "readOnly": False,
                "deviceConfig": {
                    "delay": 1,
                    "labels": "pinz",
                    "limits": [-50, 50],
                    "name": "pinz",
                    "speed": 100,
                    "tolerance": 0.01,
                    "update_frequency": 400,
                },
                "readoutPriority": "baseline",
                "onFailure": "retry",
            },
            "transd": {
                "deviceClass": "SimMonitor",
                "deviceTags": ["beamline"],
                "enabled": True,
                "readOnly": False,
                "deviceConfig": {"labels": "transd", "name": "transd", "tolerance": 0.5},
                "readoutPriority": "monitored",
                "onFailure": "retry",
            },
        }
        mock_open().write.assert_called_once_with(yaml.dump(out_data))


@pytest.fixture
def config_helper():
    connector = mock.MagicMock()
    config_helper_inst = ConfigHelper(connector)
    with mock.patch.object(config_helper_inst, "wait_for_config_reply"):
        with mock.patch.object(config_helper_inst, "wait_for_service_response"):
            yield config_helper_inst


def test_send_config_request_raises_with_empty_config(config_helper):
    with pytest.raises(DeviceConfigError):
        config_helper.send_config_request(action="update")
        config_helper.wait_for_config_reply.assert_called_once_with(mock.ANY)


def test_send_config_request(config_helper):
    config_helper.send_config_request(action="update", config={"test": "test"})
    config_helper.wait_for_config_reply.return_value = messages.RequestResponseMessage(
        accepted=True, message={"msg": "test"}
    )
    config_helper.wait_for_config_reply.assert_called_once_with(mock.ANY)
    config_helper.wait_for_service_response.assert_called_once_with(mock.ANY)


def test_send_config_request_raises_for_rejected_update(config_helper):
    config_helper.wait_for_config_reply.return_value = messages.RequestResponseMessage(
        accepted=False, message={"msg": "test"}
    )
    with pytest.raises(DeviceConfigError):
        config_helper.send_config_request(action="update", config={"test": "test"})
        config_helper.wait_for_config_reply.assert_called_once_with(mock.ANY)


def test_wait_for_config_reply():
    connector = mock.MagicMock()
    config_helper = ConfigHelper(connector)
    connector.get.return_value = messages.RequestResponseMessage(
        accepted=True, message={"msg": "test"}
    )

    res = config_helper.wait_for_config_reply("test")
    assert res == messages.RequestResponseMessage(accepted=True, message={"msg": "test"})


def test_wait_for_config_raises_timeout():
    connector = mock.MagicMock()
    config_helper = ConfigHelper(connector)
    connector.get.return_value = None

    with pytest.raises(DeviceConfigError):
        config_helper.wait_for_config_reply("test", timeout=0.3)


def test_wait_for_service_response():
    connector = mock.MagicMock()
    config_helper = ConfigHelper(connector)
    connector.lrange.side_effect = [
        [],
        [
            messages.ServiceResponseMessage(
                response={"service": "DeviceServer"}, metadata={"RID": "test"}
            ),
            messages.ServiceResponseMessage(
                response={"service": "ScanServer"}, metadata={"RID": "test"}
            ),
        ],
    ]

    config_helper.wait_for_service_response("test", timeout=0.3)


def test_wait_for_service_response_raises_timeout():
    connector = mock.MagicMock()
    config_helper = ConfigHelper(connector)
    connector.lrange.return_value = []

    with pytest.raises(DeviceConfigError):
        config_helper.wait_for_service_response("test", timeout=0.3)


def test_update_base_path_recovery():
    with mock.patch("bec_lib.bec_service.SERVICE_CONFIG") as mock_service_config:
        with mock.patch("bec_lib.config_helper.DeviceConfigWriter") as mock_device_config_writer:
            config = {"log_writer": {"base_path": "./"}}
            mock_service_config.config = {"service_config": config}
            connector = mock.MagicMock()
            config_helper = ConfigHelper(connector)
            dir_path = os.path.join(
                config["log_writer"]["base_path"], "device_configs/recovery_configs"
            )
            instance = mock_device_config_writer.get_recovery_directory
            instance.return_value = dir_path
            config_helper._update_base_path_recovery()
            assert mock_device_config_writer.call_args == mock.call(config["log_writer"])
            mock_service_config.config = {"service_config": {}}
            with pytest.raises(ServiceConfigError):
                config_helper._update_base_path_recovery()



import builtins

from rich import box
from rich.table import Table

from bec_ipython_client.beamline_mixin import BeamlineShowInfo


class SLSInfo(BeamlineShowInfo):
    def _get_sls_info(self) -> dict:
        dev = builtins.__dict__.get("dev")
        info = dev.sls_info.read(cached=True)
        return info

    def show(self):
        """Display information about the current SLS status"""

        console = self._get_console()

        table = Table(title="SLS Info", box=box.SQUARE)
        table.add_column("Key", justify="left")
        table.add_column("Value", justify="left")

        info = self._get_sls_info()

        self._add_machine_status(table, info)
        self._add_injection_mode(table, info)
        self._add_ring_current(table, info)
        self._add_current_threshold(table, info)
        self._add_current_deadband(table, info)
        self._add_filling_pattern(table, info)
        self._add_filling_lifetime(table, info)
        self._add_ofb_mode(table, info)
        self._add_fofb(table, info)
        self._add_crane_usage(table, info)
        console.print(table)

    def _add_machine_status(self, table, info):
        val = self._get_info_val(info, "sls_info_machine_status")
        if val not in ["Light Available", "Light-Available"]:
            return table.add_row("Machine status", val, style=self.ALARM_STYLE)
        return table.add_row("Machine status", val, style=self.DEFAULT_STYLE)

    def _add_injection_mode(self, table, info):
        val = self._get_info_val(info, "sls_info_injection_mode")
        if val not in ["TOP-UP", "FREQ-REFILL"]:
            return table.add_row("Injection mode", val, style=self.ALARM_STYLE)
        return table.add_row("Injection mode", val, style=self.DEFAULT_STYLE)

    def _add_current_threshold(self, table, info):
        val = info["sls_info_current_threshold"]["value"]
        if val < 350:
            return table.add_row("Current threshold", str(val), style=self.ALARM_STYLE)
        return table.add_row("Current threshold", str(val), style=self.DEFAULT_STYLE)

    def _add_current_deadband(self, table, info):
        val = info["sls_info_current_deadband"]["value"]
        if val > 2:
            return table.add_row("Current deadband", str(val), style=self.ALARM_STYLE)
        return table.add_row("Current deadband", str(val), style=self.DEFAULT_STYLE)

    def _add_filling_pattern(self, table, info):
        val = self._get_info_val(info, "sls_info_filling_pattern")
        return table.add_row("Filling pattern", val, style=self.DEFAULT_STYLE)

    def _add_filling_lifetime(self, table, info):
        val = info["sls_info_filling_life_time"]["value"]
        return table.add_row("SLS filling lifetime", f"{val:.2f} h", style=self.DEFAULT_STYLE)

    def _add_ofb_mode(self, table, info):
        val = self._get_info_val(info, "sls_info_orbit_feedback_mode")
        if val not in ["fast"]:
            return table.add_row("Orbit feedback mode", val, style=self.ALARM_STYLE)
        return table.add_row("Orbit feedback mode", val, style=self.DEFAULT_STYLE)

    def _add_fofb(self, table, info):
        val = self._get_info_val(info, "sls_info_fast_orbit_feedback")
        if val not in ["running"]:
            return table.add_row("Fast orbit feedback", val, style=self.ALARM_STYLE)
        return table.add_row("Fast orbit feedback", val, style=self.DEFAULT_STYLE)

    def _add_ring_current(self, table, info):
        val = info["sls_info_ring_current"]["value"]
        if val < 300:
            return table.add_row("Ring current", f"{val:.3f} mA", style=self.ALARM_STYLE)
        return table.add_row("Ring current", f"{val:.3f} mA", style=self.DEFAULT_STYLE)

    def _add_crane_usage(self, table, info):
        val = self._get_info_val(info, "sls_info_crane_usage")
        return table.add_row("SLS crane usage", val, style=self.DEFAULT_STYLE)


class OperatorInfo(BeamlineShowInfo):
    def _get_operator_messages(self) -> dict:
        dev = builtins.__dict__.get("dev")
        info = dev.sls_operator.read(cached=True)
        if set(info.keys()) != {f"sls_operator_messages_message{i}" for i in range(1, 6)}:
            ValueError("Unexpected data structure for sls operator messages.")
        return info

    def show(self):
        """Display information about the current SLS status"""

        console = self._get_console()
        table = Table(title="SLS Operator messages", box=box.SQUARE)
        table.add_column("Message", justify="left")
        table.add_column("Time", justify="left")

        info = self._get_operator_messages()

        for i in range(1, 6):
            msg = info[f"sls_operator_messages_message{i}"]["value"]
            date = info[f"sls_operator_date_message{i}"]["value"]
            if msg:
                table.add_row(msg, date)
        if table.row_count == 0:
            table.add_row("No messages available", "")
        console.print(table)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

import numpy as np

from bec_ipython_client.progressbar import DeviceProgressBar
from bec_lib import DeviceManagerBase, MessageEndpoints, messages

from .utils import LiveUpdatesBase, check_alarms

if TYPE_CHECKING:
    from bec_lib.client import BECClient


class ReadbackDataMixin:
    def __init__(self, device_manager: DeviceManagerBase, devices: list) -> None:
        """Mixin to get the current device values and request-done messages.

        Args:
            device_manager (DeviceManagerBase): device manager
            devices (list): list of devices to monitor
        """
        self.device_manager = device_manager
        self.devices = devices

    def get_device_values(self) -> list:
        """get the current device values

        Returns:
            list: list of device values
        """
        values = []
        for dev in self.devices:
            val = self.device_manager.devices[dev].read(cached=True)
            if not val:
                values.append(np.nan)
                continue
            # pylint: disable=protected-access
            hints = self.device_manager.devices[dev]._hints
            # if we have hints, use them to get the value, otherwise just use the first value
            if hints:
                values.append(val.get(hints[0]).get("value"))
            else:
                values.append(val.get(list(val.keys())[0]).get("value"))
        return values

    def get_request_done_msgs(self):
        """get all request-done messages"""
        pipe = self.device_manager.connector.pipeline()
        for dev in self.devices:
            self.device_manager.connector.get(MessageEndpoints.device_req_status(dev), pipe)
        return self.device_manager.connector.execute_pipeline(pipe)

    def wait_for_RID(self, request: messages.ScanQueueMessage) -> None:
        """wait for the readback's metadata to match the request ID

        Args:
            request (messages.ScanQueueMessage): request message
        """
        while True:
            msgs = [
                self.device_manager.connector.get(MessageEndpoints.device_readback(dev))
                for dev in self.devices
            ]
            if all(msg.metadata.get("RID") == request.metadata["RID"] for msg in msgs if msg):
                break
            check_alarms(self.device_manager.parent)


class LiveUpdatesReadbackProgressbar(LiveUpdatesBase):
    """Live feedback on motor movements using a progressbar.

    Args:
        dm (DeviceManagerBase): device_manager
        request (ScanQueueMessage): request that should be monitored

    """

    def __init__(
        self,
        bec: BECClient,
        report_instruction: list = None,
        request: messages.ScanQueueMessage = None,
        callbacks: list[Callable] = None,
    ) -> None:
        super().__init__(
            bec, report_instruction=report_instruction, request=request, callbacks=callbacks
        )
        if report_instruction:
            self.devices = report_instruction["readback"]["devices"]
        else:
            self.devices = list(request.content["parameter"]["args"].keys())

    def core(self):
        """core function to monitor the device values and update the progressbar accordingly."""
        data_source = ReadbackDataMixin(self.bec.device_manager, self.devices)
        start_values = data_source.get_device_values()
        self.wait_for_request_acceptance()
        data_source.wait_for_RID(self.request)
        if self.report_instruction:
            self.devices = self.report_instruction["readback"]["devices"]
            target_values = self.report_instruction["readback"]["end"]

            start_instr = self.report_instruction["readback"].get("start")
            if start_instr:
                start_values = self.report_instruction["readback"]["start"]
            data_source = ReadbackDataMixin(self.bec.device_manager, self.devices)
        else:
            target_values = [
                x for xs in self.request.content["parameter"]["args"].values() for x in xs
            ]
            if self.request.content["parameter"]["kwargs"].get("relative"):
                target_values = np.asarray(target_values) + np.asarray(start_values)

        with DeviceProgressBar(
            self.devices, start_values=start_values, target_values=target_values
        ) as progress:
            req_done = False
            while not progress.finished or not req_done:
                check_alarms(self.bec)

                values = data_source.get_device_values()
                progress.update(values=values)
                self._print_client_msgs_asap()

                msgs = data_source.get_request_done_msgs()
                request_ids = [
                    msg.metadata["RID"] if (msg and msg.metadata.get("RID")) else None
                    for msg in msgs
                ]
                if set(request_ids) != set([self.request.metadata["RID"]]):
                    progress.sleep()
                    continue

                req_done = True
                for dev, msg in zip(self.devices, msgs):
                    if not msg:
                        continue
                    if msg.content.get("success", False):
                        progress.set_finished(dev)
                # pylint: disable=protected-access
                progress._progress.refresh()
        # TODO #286
        # self._print_client_msgs_all()

    def run(self):
        """run the progressbar."""
        self.core()



from bec_lib.scan_report import ScanReport

# pylint:disable=undefined-variable
# pylint: disable=too-many-arguments


def umv(*args) -> ScanReport:
    """Updated absolute move (i.e. blocking) for one or more devices.

    Returns:
        ScanReport: Status object.

    Examples:
        >>> umv(dev.samx, 1)
        >>> umv(dev.samx, 1, dev.samy, 2)
    """
    return scans.umv(*args, relative=False)


def umvr(*args) -> ScanReport:
    """Updated relative move (i.e. blocking) for one or more devices.

    Returns:
        ScanReport: Status object.

    Examples:
        >>> umvr(dev.samx, 1)
        >>> umvr(dev.samx, 1, dev.samy, 2)
    """
    return scans.umv(*args, relative=True)


def mv(*args) -> ScanReport:
    """Absolute move for one or more devices.

    Returns:
        ScanReport: Status object.

    Examples:
        >>> mv(dev.samx, 1)
        >>> mv(dev.samx, 1, dev.samy, 2)
    """
    return scans.mv(*args, relative=False)


def mvr(*args) -> ScanReport:
    """Relative move for one or more devices.

    Returns:
        ScanReport: Status object.

    Examples:
        >>> mvr(dev.samx, 1)
        >>> mvr(dev.samx, 1, dev.samy, 2)
    """
    return scans.mv(*args, relative=True)



import collections
from unittest import mock

import pytest

from bec_ipython_client.callbacks.move_device import (
    LiveUpdatesReadbackProgressbar,
    ReadbackDataMixin,
)
from bec_lib import MessageEndpoints, messages


@pytest.fixture
def readback_data_mixin(bec_client_mock):
    with mock.patch.object(bec_client_mock.device_manager, "connector"):
        yield ReadbackDataMixin(bec_client_mock.device_manager, ["samx", "samy"])


def test_move_callback(bec_client_mock):
    client = bec_client_mock
    request = messages.ScanQueueMessage(
        scan_type="umv",
        parameter={"args": {"samx": [10]}, "kwargs": {"relative": True}},
        metadata={"RID": "something"},
    )
    readback = collections.deque()
    readback.extend([[-10], [0], [10]])

    def mock_readback(*args):
        if len(readback) > 1:
            return readback.popleft()
        return readback[0]

    req_done = collections.deque()
    msg_acc = messages.DeviceReqStatusMessage(
        device="samx", success=True, metadata={"RID": "something"}
    )
    req_done.extend([[None], [None], [None], [msg_acc]])

    def mock_req_msg(*args):
        if len(req_done) > 1:
            return req_done.popleft()
        return req_done[0]

    with mock.patch("bec_ipython_client.callbacks.move_device.check_alarms") as check_alarms_mock:
        with mock.patch.object(ReadbackDataMixin, "wait_for_RID"):
            with mock.patch.object(LiveUpdatesReadbackProgressbar, "wait_for_request_acceptance"):
                with mock.patch.object(
                    LiveUpdatesReadbackProgressbar, "_print_client_msgs_asap"
                ) as mock_client_msgs:
                    with mock.patch.object(ReadbackDataMixin, "get_device_values", mock_readback):
                        with mock.patch.object(
                            ReadbackDataMixin, "get_request_done_msgs", mock_req_msg
                        ):
                            LiveUpdatesReadbackProgressbar(bec=client, request=request).run()
                            assert mock_client_msgs.called


def test_move_callback_with_report_instruction(bec_client_mock):
    client = bec_client_mock
    request = messages.ScanQueueMessage(
        scan_type="umv",
        parameter={"args": {"samx": [10]}, "kwargs": {"relative": True}},
        metadata={"RID": "something"},
    )
    readback = collections.deque()
    readback.extend([[-10], [0], [10]])
    report_instruction = {
        "readback": {"RID": "something", "devices": ["samx"], "start": [0], "end": [10]}
    }

    def mock_readback(*args):
        if len(readback) > 1:
            return readback.popleft()
        return readback[0]

    req_done = collections.deque()
    msg_acc = messages.DeviceReqStatusMessage(
        device="samx", success=True, metadata={"RID": "something"}
    )
    req_done.extend([[None], [None], [None], [msg_acc]])

    def mock_req_msg(*args):
        if len(req_done) > 1:
            return req_done.popleft()
        return req_done[0]

    with mock.patch("bec_ipython_client.callbacks.move_device.check_alarms") as check_alarms_mock:
        with mock.patch.object(ReadbackDataMixin, "wait_for_RID"):
            with mock.patch.object(LiveUpdatesReadbackProgressbar, "wait_for_request_acceptance"):
                with mock.patch.object(
                    LiveUpdatesReadbackProgressbar, "_print_client_msgs_asap"
                ) as mock_client_msgs:
                    with mock.patch.object(ReadbackDataMixin, "get_device_values", mock_readback):
                        with mock.patch.object(
                            ReadbackDataMixin, "get_request_done_msgs", mock_req_msg
                        ):
                            LiveUpdatesReadbackProgressbar(
                                bec=client, report_instruction=report_instruction, request=request
                            ).run()


def test_readback_data_mixin(readback_data_mixin):
    readback_data_mixin.device_manager.connector.get.side_effect = [
        messages.DeviceMessage(
            signals={"samx": {"value": 10}, "samx_setpoint": {"value": 20}},
            metadata={"device": "samx"},
        ),
        messages.DeviceMessage(
            signals={"samy": {"value": 10}, "samy_setpoint": {"value": 20}},
            metadata={"device": "samy"},
        ),
    ]
    res = readback_data_mixin.get_device_values()
    assert res == [10, 10]


def test_readback_data_mixin_multiple_hints(readback_data_mixin):
    readback_data_mixin.device_manager.devices.samx._info["hints"]["fields"] = [
        "samx_setpoint",
        "samx",
    ]
    readback_data_mixin.device_manager.connector.get.side_effect = [
        messages.DeviceMessage(
            signals={"samx": {"value": 10}, "samx_setpoint": {"value": 20}},
            metadata={"device": "samx"},
        ),
        messages.DeviceMessage(
            signals={"samy": {"value": 10}, "samy_setpoint": {"value": 20}},
            metadata={"device": "samy"},
        ),
    ]
    res = readback_data_mixin.get_device_values()
    assert res == [20, 10]


def test_readback_data_mixin_multiple_no_hints(readback_data_mixin):
    readback_data_mixin.device_manager.devices.samx._info["hints"]["fields"] = []
    readback_data_mixin.device_manager.connector.get.side_effect = [
        messages.DeviceMessage(
            signals={"samx": {"value": 10}, "samx_setpoint": {"value": 20}},
            metadata={"device": "samx"},
        ),
        messages.DeviceMessage(
            signals={"samy": {"value": 10}, "samy_setpoint": {"value": 20}},
            metadata={"device": "samy"},
        ),
    ]
    res = readback_data_mixin.get_device_values()
    assert res == [10, 10]


def test_get_request_done_msgs(readback_data_mixin):
    res = readback_data_mixin.get_request_done_msgs()
    readback_data_mixin.device_manager.connector.pipeline.assert_called_once()
    assert (
        mock.call(
            MessageEndpoints.device_req_status("samx"),
            readback_data_mixin.device_manager.connector.pipeline.return_value,
        )
        in readback_data_mixin.device_manager.connector.get.call_args_list
    )
    assert (
        mock.call(
            MessageEndpoints.device_req_status("samy"),
            readback_data_mixin.device_manager.connector.pipeline.return_value,
        )
        in readback_data_mixin.device_manager.connector.get.call_args_list
    )



from abc import abstractmethod

from rich import style
from rich.console import Console


class BeamlineShowInfo:
    DEFAULT_STYLE = style.Style(color="green")
    ALARM_STYLE = style.Style(color="red", bold=True)

    @staticmethod
    def _get_info_val(info, entry):
        return str(info[entry]["value"])

    def _get_console(self) -> Console:
        return Console()

    @abstractmethod
    def show(self):
        """Display the info"""


class BeamlineMixin:
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._bl_calls = []

    def bl_show_all(self):
        """Display general information about the SLS and the current status of the beamline"""
        for call in self._bl_calls:
            call.show()

    def _bl_info_register(self, bl_info: type[BeamlineShowInfo]) -> None:
        """Register a beamline info class

        Args:
            bl_info (Type[BeamlineShowInfo]): Class whose instance ought to be registered.
        """
        bli = bl_info()
        self._bl_calls.append(bli)
